import numpy as np
from veux.model import Model
import matplotlib.pyplot as plt
import matplotlib.tri as tri

def _plot_grid(x,y, ax=None, **kwargs):
    ax = ax or plt.gca()
    segs1 = np.stack((x,y), axis=2)
    segs2 = segs1.transpose(1,0,2)
    ax.add_collection(LineCollection(segs1, **kwargs))
    ax.add_collection(LineCollection(segs2, **kwargs))

    ax.autoscale()


# converts quad elements into tri elements
def _quads_to_tris(quads):
    tris = [
        [None for j in range(3)] for i in range(2*len(quads))
    ]
    for i in range(len(quads)):
        j = 2*i
        tris[j][0] = quads[i][0]
        tris[j][1] = quads[i][1]
        tris[j][2] = quads[i][2]
        tris[j + 1][0] = quads[i][2]
        tris[j + 1][1] = quads[i][3]
        tris[j + 1][2] = quads[i][0]
    return tris

class PlaneModel(Model):
    ndm = 2
    def __init__(self, mesh, ndf=2):
        self.recs = []
        self.tris = []
        self.ndf = ndf

        if isinstance(mesh, tuple):
            nodes, elems = mesh
            if isinstance(nodes, dict):
                self.nodes = nodes 
            else:
                self.nodes = {i: list(coord) for i, coord in enumerate(nodes)}
            self.recs = [
    #            tuple(i-1 for i in elem) for elem in elems.values() if len(elem) == 4
                tuple(i   for i in elem) for elem in elems if len(elem) == 4
            ]
            self.tris = [
                tuple(i for i in elem) for elem in elems if len(elem) == 3
            ]

        else:
            # assume mesh is a meshio object
            self.nodes = {i: list(coord)[:2] for i, coord in enumerate(mesh.points)}
            for blk in mesh.cells:
                if blk.type == "triangle":
                    self.tris = self.tris + [
                        tuple(int(i) for i in elem) for elem in blk.data
                    ]
                elif blk.type == "quad":
                    self.recs = self.recs + [
                        tuple(int(i) for i in elem) for elem in blk.data
                    ]

    def frame_orientation(self, tag):
        return None


    def iter_node_tags(self):
        for tag in self.nodes:
            yield tag 

    def iter_cell_tags(self):
        for tag in range(len(self.tris)+len(self.recs)):
            yield tag

    def cell_matches(self, tag, type=None):
        if type == "plane":
            return True 
        else:
            return False

    def cell_exterior(self, tag=None):
        if tag is None:
            return self.tris + self.recs 
        elif tag < len(self.tris):
            return self.tris[tag]
        else:
            return self.recs[tag]

    def node_position(self, tag=None, state=None):
        xyz = np.zeros((len(self.nodes), 3))
        xyz[:,[0,2]] = list(self.nodes.values())
        return xyz

    def cell_triangles(self, tag=None):
        if tag is None:
            node_tag_to_index = {tag: i for i, tag in enumerate(self.nodes.keys())}
            return [
                    tuple(node_tag_to_index[tag] for tag in elem)
                    for elem in self.tris + _quads_to_tris(self.recs)
            ]
        elif tag < len(self.tris):
            return self.tris[tag]
        else:
            return self.recs[tag]


class PlaneArtist:
    def __init__(self, model, ax=None, **kwds):
        if ax is None:
            _,ax = plt.subplots()
        self.ax = ax

        self.model = model

    def _draw_nodes(self, nodes):
        self.ax.scatter(*zip(*nodes.values()))
        for k,v in nodes.items():
            self.ax.annotate(k, v)

    def draw_outlines(self, **kwds):
        ax = self.ax
        # TODO:
        nodes = self.model.nodes

        for element in self.model.cell_exterior():
            x = [nodes[element[i]][0] for i in range(len(element))]
            y = [nodes[element[i]][1] for i in range(len(element))]
            ax.fill(x, y, edgecolor='black', ls="-", lw=0.5, fill=False)


    def draw_surfaces(self, field=None, show_scale=False):
        ax = self.ax 

        #
        # Plot solution contours
        #
        nodes_x, _, nodes_y = self.model.node_position().T

        triangles = self.model.cell_triangles()

        # create an unstructured triangular grid instance
        triangulation = tri.Triangulation(nodes_x, nodes_y, triangles)
        contours = \
            ax.tricontourf(triangulation, field, cmap="twilight", alpha=0.5)

        if show_scale:
            plt.colorbar(contours, ax=ax)

    def draw(self):
        self.ax.axis('equal')

    def show(self):
        plt.show()

def render(mesh, field=None, ax=None,
         # mesh options
         show_edges=True,
         # contour options
         show_scale=True
    ):
    artist = PlaneArtist(PlaneModel(mesh))

    #
    # plot the finite element mesh
    #
    if show_edges:
        artist.draw_outlines()

    if field is not None:
        artist.draw_surfaces(field=field, show_scale=show_scale)

    artist.draw() 
    return artist



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Claudio Perez
#
from .canvas import Canvas, NodeStyle, MeshStyle, LineStyle
import numpy as np

class PlotlyCanvas(Canvas):
    vertical = 3

    def __init__(self, config=None):
        self.data = []
        self.config = config
        self.annotations = []
        self._vertices = []

    def plot_nodes(self, vertices, label = None, style=None, data=None, rotations=None):
        name = label or "nodes"
        x,y,z = vertices.T

        trace = {
                "name": name,
                "x": x, "y": y, "z": z,
                "type": "scatter3d","mode": "markers",
#               "hovertemplate": "<br>".join(f"{k}: %{{customdata[{v}]}}" for v,k in enumerate(keys)),
#               "customdata": data,
                "marker": {
                    "symbol": "square",
                    "size": 3*style.scale,
                    "color": style.color
                    #**self.config["objects"]["nodes"]["default"]
                },
                "showlegend": False
        }
        self.data.append(trace)

    def plot_lines(self, vertices, indices=None, label=None, style=None)->list:
        lines=[]
        if indices is not None:
            for idx in indices:
                lines.extend(self.plot_lines(vertices[idx], label=label, style=style))
            return lines

        if style is None:
            style = LineStyle(color="#808080")


        x,y,z = vertices.T
        data = {
            "name": label if label is not None else "",
            "type": "scatter3d",
            "mode": "lines",
            "projection": dict(
                z=dict(show=True),
            ),
            "x": x, "y": y, "z": z,
            "line": {"color": style.color, "width": style.width},
            "hoverinfo":"skip"
        }
        self.data.append(data)
        return lines


    def plot_hover(self, vertices, data=None, keys=None, text=None, html=None, style=None, label=None):
            if isinstance(data, dict):
                keys  = list(map(str, data.keys()))
                data  = [list(map(str, data.values())) for _ in range(len(vertices))]

            elif isinstance(data, list):
                assert keys is not None

            x, y, z = vertices.T

            self.data.append({
                "name": label,
                "x": x, "y": y, "z": z,
                "type": "scatter3d", "mode": "markers", # "lines", #
                "hovertemplate": "<br>".join(f"{k}: %{{customdata[{v}]}}" for v,k in enumerate(keys)),
                "customdata": data,
                "opacity": 0
            })

    def annotate(self, text, coords, **kwds):
        self.annotations.append({
            "text": text,
            "showarrow": False,
            "xanchor": "left",
            "yshift": -10,
           #"yshift":  10, # For hockling
            "xshift":   5,
            "font": {"color": "black", "size": 48},
            "x": coords[0], "y": coords[1], "z": coords[2],
            **kwds
        })

    def plot_vectors(self, locs, vecs, **kwds):

        ne = vecs.shape[0]
        for j in range(3):
            X = np.zeros((ne*3, 3))*np.nan
            for i in range(j,ne,3):
                X[i*3,:] = locs[i]
                X[i*3+1,:] = locs[i] + vecs[i]

            color = kwds.get("color", ("red", "blue", "green")[j])

            # _label = label if label is not None else ""
            label = kwds.get("label", "")
            if isinstance(label, list):
                label = label[j]
            self.data.append({
                "name": label,
                "type": "scatter3d",
                "mode": "lines",
                "x": X.T[0], "y": X.T[1], "z": X.T[2],
                "line": {"color": color, "width": 4},
                "hoverinfo":"skip",
                "showlegend": False
            })

    def plot_mesh(self, vertices, triangles, style=None, local_coords=None)->tuple:
        if style is None:
            style = MeshStyle()

        if isinstance(vertices, int):
            x, y, z = self._vertices[vertices]
            point_access = vertices
        else:
            x,y,z = zip(*vertices)
            self._vertices.append(vertices)
            point_access = len(self._vertices) - 1

        i,j,k = zip(*triangles)
        self.data.append({
            #"name": label if label is not None else "",
            "type": "mesh3d",
            "x": x, "y": y, "z": z,
            "i": i, "j": j, "k": k,
            "hoverinfo":"skip",
#           "lighting": dict(ambient=0.9, roughness=0.5, specular=2),
#           "lighting": dict(ambient=0.4, diffuse=0.5, roughness = 0.9, specular=0.6, fresnel=0.2),
            "opacity": style.alpha,
            "color":   style.color,
        })
        return point_access, None

    def build(self):
        opts = self.config
        show_objects = set()
#       show_axis = "axis" in opts["show_objects"]
        show_grid = False #"grid" in opts["show_objects"]
        import plotly.graph_objects as go
        fig = go.Figure(dict(
                data=self.data,
                rendermode='webgl',
                layout=go.Layout(
#                 paper_bgcolor='white',
                  scene=dict(aspectmode="data",
                         xaxis = {"showspikes": "tick" in show_objects, "nticks": 0,
                              "showbackground": show_grid, "backgroundcolor": "white",
                              "showticklabels": "tick" in show_objects,
                              "gridcolor": "gray" if show_grid else None
                         },
                         yaxis = {"showspikes": "tick" in show_objects, "nticks": 0,
                              "showbackground": show_grid, "backgroundcolor": "white",
                              "showticklabels": "tick" in show_objects,
                              "gridcolor": "gray" if show_grid else None
                         },
                         zaxis = {"showspikes": "tick" in show_objects, "nticks": 0,
                              "showbackground": show_grid, "backgroundcolor": "white",
                              "showticklabels": "tick" in show_objects,
                              "gridcolor": "gray" if show_grid else None
                         },
                     # LaTeX is not currently rendered in 3D, see:
                     # https://github.com/plotly/plotly.js/issues/608
#                    xaxis_title = r"$\mathbf{E}_1$",
                     xaxis_title = "", yaxis_title="", zaxis_title="",
                     xaxis_visible =  "x" in show_objects,#show_axis,
                     yaxis_visible =  "y" in show_objects,#show_axis,
                     zaxis_visible =  "z" in show_objects,#show_axis,
                     camera=dict(
                         projection={"type": opts["camera"]["projection"]}
                     ),
                     annotations=self.annotations,
                  ),
                  margin=dict(l=0,r=0,b=0,t=25),
#                 showlegend=("legend" in show_objects)
                )
            ))
        fig.update(layout_coloraxis_showscale=False)

        fig.update_traces(selector=dict(type='scatter3d', mode='lines'),
                          projection=dict(
                              z=dict(show=True),
                          )
        )
        self.fig = fig
        return self

    def to_html(self):
        import plotly
        #import plotly.offline
        #plotly.offline.plot(data, include_plotlyjs=False, output_type='div')
        options = {
          "include_plotlyjs": True,
          "include_mathjax" : "cdn",
          "full_html"       : True
        }
        return plotly.io.to_html(self.fig,
                                 div_id=str(id(self)),
                                 **options)

    def write(self, filename, format=None, **options):
        if "html" in filename:
            with open(filename,"w+") as f:
                f.write(self.to_html())

        elif "png" in filename:
            self.fig.write_image(filename, width=1920, height=1080)

        elif "json" in filename:
            with open(filename,"w+") as f:
                self.fig.write_json(f)

    def make_hover_data(self, data, ln=None):
        if ln is None:
            items = np.array([d.values for d in data])
            keys = data[0].keys()
        else:
            items = np.array([list(data.values())]*ln)
            keys = data.keys()
        return {
            "hovertemplate": "<br>".join(f"{k}: %{{customdata[{v}]}}" for v,k in enumerate(keys)),
            "customdata": list(items),
        }




import math
import pygltflib
from pygltflib import GLTF2, Scene, Node, Mesh, Material
from pygltflib import (
    Buffer, BufferView, Accessor, Animation, 
    AnimationSampler, AnimationChannel, AnimationChannelTarget
)
from pygltflib import ELEMENT_ARRAY_BUFFER, ARRAY_BUFFER
from pygltflib import FLOAT, UNSIGNED_SHORT
from pygltflib import Primitive, Skin

import struct

# Helper to multiply two quaternions (x, y, z, w).
def quaternion_multiply(q1, q2):
    x1, y1, z1, w1 = q1
    x2, y2, z2, w2 = q2
    x = w1*x2 + x1*w2 + y1*z2 - z1*y2
    y = w1*y2 + y1*w2 + z1*x2 - x1*z2
    z = w1*z2 + z1*w2 + x1*y2 - y1*x2
    w = w1*w2 - x1*x2 - y1*y2 - z1*z2
    return (x, y, z, w)

# -----------------------------
# 1) Define geometry for a simple line
#    Make it a bit longer so the swing is more obvious.
# -----------------------------
positions = [
    0.0,  0.0,  0.0,  # pivot (origin)
    0.0, -2.0,  0.0   # longer pendulum tip
]
# Indices for a line segment
indices = [0, 1]

positions_bytes = b"".join([struct.pack("<f", p) for p in positions])
indices_bytes   = b"".join([struct.pack("<H", i) for i in indices])
combined_buffer = positions_bytes + indices_bytes

position_buffer_view = BufferView(
    buffer=0,
    byteOffset=0,
    byteLength=len(positions_bytes),
    target=ARRAY_BUFFER
)
index_buffer_view = BufferView(
    buffer=0,
    byteOffset=len(positions_bytes),
    byteLength=len(indices_bytes),
    target=ELEMENT_ARRAY_BUFFER
)

position_accessor = Accessor(
    bufferView=0,
    byteOffset=0,
    componentType=FLOAT,
    count=len(positions)//3,
    type="VEC3",
    min=[0.0, -2.0,  0.0],
    max=[0.0,  0.0,  0.0]
)
index_accessor = Accessor(
    bufferView=1,
    byteOffset=0,
    componentType=UNSIGNED_SHORT,
    count=len(indices),
    type="SCALAR",
    min=[0],
    max=[1]
)

mesh_primitive = Primitive(
    attributes={"POSITION": 0},
    indices=1,
    mode=pygltflib.LINES
)
mesh = Mesh(primitives=[mesh_primitive])

pendulum_node = Node(
    mesh=0,
    name="PendulumNode"
)

# -----------------------------
# 5) Build a more pronounced 3D motion animation.
# -----------------------------
period = 2.0
num_samples = 10

times = []
rotations = []

# Larger amplitude around Z, plus a small amplitude around X.
amplitude_degs_z = 45  # ±45 degrees around Z
amplitude_degs_x = 10  # ±10 degrees around X

omega = 2 * math.pi / period
amplitude_rads_z = math.radians(amplitude_degs_z)
amplitude_rads_x = math.radians(amplitude_degs_x)

for i in range(num_samples):
    t = i * (period / (num_samples - 1))
    # Angle around Z
    angle_z = amplitude_rads_z * math.cos(omega * t)
    # Angle around X
    angle_x = amplitude_rads_x * math.sin(omega * t)
    
    # Convert each to a quaternion:
    #   qZ for rotation around Z
    #   qX for rotation around X
    half_z = angle_z / 2
    half_x = angle_x / 2
    
    sin_z = math.sin(half_z)
    cos_z = math.cos(half_z)
    # rotation around Z = (0, 0, sin(θ/2), cos(θ/2))
    qZ = (0.0, 0.0, sin_z, cos_z)
    
    sin_x = math.sin(half_x)
    cos_x = math.cos(half_x)
    # rotation around X = (sin(θ/2), 0, 0, cos(θ/2))
    qX = (sin_x, 0.0, 0.0, cos_x)
    
    # Combine them: qTotal = qZ * qX
    qTotal = quaternion_multiply(qZ, qX)
    
    times.append(t)
    rotations.append(qTotal)

time_bytes = b"".join([struct.pack("<f", t) for t in times])
rotation_bytes = b"".join([struct.pack("<4f", *r) for r in rotations])

animation_time_buffer_view = BufferView(
    buffer=0,
    byteOffset=len(combined_buffer),
    byteLength=len(time_bytes),
    target=None
)
animation_rotation_buffer_view = BufferView(
    buffer=0,
    byteOffset=len(combined_buffer) + len(time_bytes),
    byteLength=len(rotation_bytes),
    target=None
)

time_accessor = Accessor(
    bufferView=2,
    byteOffset=0,
    componentType=FLOAT,
    count=len(times),
    type="SCALAR",
    min=[times[0]],
    max=[times[-1]]
)
rotation_accessor = Accessor(
    bufferView=3,
    byteOffset=0,
    componentType=FLOAT,
    count=len(rotations),
    type="VEC4"
)

animation_sampler = AnimationSampler(
    input=2,
    output=3,
    interpolation="LINEAR"
)
animation_channel = AnimationChannel(
    sampler=0,
    target=AnimationChannelTarget(
        node=0,
        path="rotation"
    )
)
animation = Animation(
    samplers=[animation_sampler],
    channels=[animation_channel],
    name="PendulumSwing"
)

gltf = GLTF2(
    scenes=[Scene(nodes=[0])],
    nodes=[pendulum_node],
    meshes=[mesh],
    animations=[animation],
    buffers=[Buffer(byteLength=len(combined_buffer) + len(time_bytes) + len(rotation_bytes))],
    bufferViews=[
        position_buffer_view, 
        index_buffer_view,
        animation_time_buffer_view,
        animation_rotation_buffer_view
    ],
    accessors=[
        position_accessor,
        index_accessor,
        time_accessor,
        rotation_accessor
    ]
)

# Store the combined binary data in the first buffer
gltf.buffers[0].uri = None
gltf._glb_data = combined_buffer + time_bytes + rotation_bytes

gltf.save("pendulum.glb")
print("pendulum.glb has been created.")




#!/usr/bin/env python3

"""
author:  Nils Olofsson
email:   nils.olovsson@gmail.com
website: http://nilsolovsson.se
date:    2021-02-28
license: MIT

https://nils-olovsson.se/articles/ear_clipping_triangulation/
https://bitbucket.org/nils_olovsson/ear_clipping_triangulation/src/master/
"""

import numpy as np

# ==============================================================================
#
# Geometry
#
# ==============================================================================

def getExtents(vertices):
    """
    Get the extents, axis aligned bounding box, of the set of vertices.
    """
    return np.array([
        np.min(vertices[:,0]), # left
        np.min(vertices[:,1]), # bottom
        np.max(vertices[:,0]), # right
        np.max(vertices[:,1])  # top
    ])

def vertexNormal(v0, v1, v2):
    """
    Compute the 2D vertex normal, for placing rendered vertex numbers.
    """
    e0 = v1 - v0
    e1 = v2 - v1
    n0 = np.array([e0[1], -e0[0]])
    n1 = np.array([e1[1], -e1[0]])
    l0 = np.linalg.norm(e0)
    l1 = np.linalg.norm(e1)
    
    #n = n0 + n1
    #n = l0*n0 + l1*n1
    n = 0.5*(1.0/l0*n0 + 1.0/l1*n1)
    n = 1.0 / np.linalg.norm(n) * n
    return n

def computeNormals(vertices):
    """
    Compute the 2D vertex normals of the polygon, for placing rendered
    vertex numbers.
    """
    n,m = vertices.shape
    normals = np.zeros(vertices.shape)
    for i in range(n):
        v0 = vertices[i-1,:]
        v1 = vertices[i,:]
        v2 = vertices[(i+1)%n,:]
        normals[i,:] = vertexNormal(v0,v1,v2)

    return normals

def angleCCW(a, b):
    """
    Counter clock wise angle (radians) from normalized 2D vectors a to b
    """
    dot = a[0]*b[0] + a[1]*b[1]
    det = a[0]*b[1] - a[1]*b[0]
    angle = np.arctan2(det, dot)
    if angle<0.0 :
        angle = 2.0*np.pi + angle
    return angle

def _is_convex(triangle):
    """
    Determine if vertex lies on the convex hull of the polygon.
    """
    vertex_prev, vertex, vertex_next = triangle
    a = vertex_prev - vertex
    b = vertex_next - vertex
    internal_angle = angleCCW(b, a)
    return internal_angle <= np.pi

def insideTriangle(triangle, p):
    """
        Determine if a vertex p is inside (or "on") a triangle made of the
        points a->b->c
        http://blackpawn.com/texts/pointinpoly/
    """
    a, b, c = triangle
    #Compute vectors
    v0 = c - a
    v1 = b - a
    v2 = p - a

    # Compute dot products
    dot00 = np.dot(v0, v0)
    dot01 = np.dot(v0, v1)
    dot02 = np.dot(v0, v2)
    dot11 = np.dot(v1, v1)
    dot12 = np.dot(v1, v2)

    # Compute barycentric coordinates
    denom = dot00*dot11 - dot01*dot01
    if abs(denom) < 1e-20:
        return True
    invDenom = 1.0 / denom
    u = (dot11*dot02 - dot01*dot12) * invDenom
    v = (dot00*dot12 - dot01*dot02) * invDenom

    # Check if point is in triangle
    return (u >= 0) and (v >= 0) and (u + v < 1)


def earcut(vertices, max_iterations=0):
    """
        Triangulation of a polygon in 2D.
        Assumption that the polygon is simple, i.e has no holes, is closed and
        has no crossings and also that it the vertex order is counter clockwise.
        https://geometrictools.com/Documentation/TriangulationByEarClipping.pdf
    """

    n, m = vertices.shape
    indices = np.zeros([n-2, 3], dtype=int)


    vertlist = DoubleLinkedList()
    for i in range(0, n):
        vertlist.append(i)

    index_counter = 0
    it_counter = 0

    # Simplest possible algorithm. Create list of indexes.
    # Find first ear vertex. Create triangle. Remove vertex from list
    # Do this while number of vertices > 2.
    node = vertlist.first

    while vertlist.size > 2 and (max_iterations<=0 or max_iterations>index_counter):
        i = node.prev.data
        j = node.data
        k = node.next.data

        vert_prev = vertices[i,:]
        vert_crnt = vertices[j,:]
        vert_next = vertices[k,:]

        is_convex = _is_convex((vert_prev, vert_crnt, vert_next))
        is_ear = True
        if is_convex:
            test_node = node.next.next
            while test_node!=node.prev and is_ear:
                vert = vertices[test_node.data,:]
                is_ear = not insideTriangle((vert_prev, vert_crnt, vert_next), vert)
                test_node = test_node.next
        else:
            is_ear = False

        if is_ear:
            indices[index_counter, :] = np.array([i, j, k], dtype=int)
            index_counter += 1
            vertlist.remove(node.data)
        it_counter += 1
        node = node.next
    indices = indices[0:index_counter, :]
    return indices, vertlist.flatten()


class Node:
    """
    Node element in a DoubleLinkedList.
    Each node in a valid list is associated with a value/data element and
    with its left and right neighbor.
    [Prev. node]<--[Node]-->[Next node]
                        |
                    [Data]
    """

    def __init__(self, data):
        self.data = data
        self.prev = None
        self.next = None

class DoubleLinkedList:
    """
    A double linked list. Each element keeps a reference to both left and
    right neighbor. This allows e.g. for easy removal of elements.
    The list is circular and is usually considered traversed when the next element
    is the same element as when when we started.
    """

    def __init__(self):
        self.first = None
        self.size  = 0

    def __str__(self):
        if self.first==None:
            return '[]'
        msg = '['
        msg += str(self.first.data)
        node = self.first.next
        while node != self.first:
            msg += ', ' + str(node.data)
            node = node.next
        msg += ']'
        return msg

    def append(self, data):
        self.size += 1
        if self.first == None:
            self.first = Node(data)
            self.first.prev = self.first
            self.first.next = self.first
            return
        node = Node(data)
        last = self.first.prev
        node.prev = last
        node.next = self.first
        last.next = node
        self.first.prev = node

    def remove(self, item):
        if self.first==None:
            return
        rmv = None
        node = self.first
        if node.data == item:
            rmv = node
        node = node.next
        while not rmv and node != self.first:
            if node.data == item:
                rmv = node
            node = node.next
        if rmv:
            nxt = rmv.next
            prv = rmv.prev
            prv.next = nxt
            nxt.prev = prv
            self.size -= 1
            if rmv == self.first:
                self.first = nxt
            if rmv == self.first:
                self.first = None
        return

    def count(self):
        if self.first==None:
            return 0
        i = 1
        node = self.first.next
        while node != self.first:
            i+=1
            node = node.next
        return i

    def flatten(self):
        if self.first==None:
            return []
        l = []
        node = self.first
        l.append(node.data)
        node = self.first.next
        while node != self.first:
            l.append(node.data)
            node = node.next
        return l




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Claudio Perez
#
"""
The glTF file format is supported by [COMSOL](https://www.comsol.com/blogs/how-to-export-and-share-your-3d-result-plots-as-gltf-files)
"""

"""
Some implementation notes:
- glTF defines +Y as up.
- glTF uses a right-handed coordinate system, that is, the cross product of +X and +Y yields +Z.
- The front of a glTF asset faces +Z.


- All angles are in radians.
- Positive rotation is counterclockwise.
- Rotations are given as quaternions stored as a tuple (x,y,z,w),
  where the w-component is the cosine of half of the rotation angle.
  For example, the quaternion [ 0.259, 0.0, 0.0, 0.966 ] describes a rotation
  about 30 degrees, around the x-axis.
- The identity rotation is (0, 0, 0, 1.0)

"""
import sys
import itertools

import numpy as np
import pygltflib
from scipy.spatial.transform import Rotation

import veux

from .canvas import Canvas, Line, Mesh, Node
from veux import utility
from veux.config import NodeStyle, MeshStyle, LineStyle, DrawStyle

GLTF_T = {
    "float32": pygltflib.FLOAT,
    "uint8":   pygltflib.UNSIGNED_BYTE,
    "uint16":  pygltflib.UNSIGNED_SHORT,
}

import numpy as np
import pygltflib

EYE3 = np.eye(3, dtype="float32")

def _append_index(lst, item):
    lst.append(item)
    return len(lst) - 1


class GltfLibCanvas(Canvas):
    vertical = 2

    def __init__(self, config=None):
        self.config = config

        #                          x, y, z, scalar
        self._rotation = [-0.7071068, 0, 0, 0.7071068]
        # equivalent rotation matrix:
        self._rotation_matrix = np.array([[1,  0, 0],
                                          [0,  0, 1],
                                          [0, -1, 0]])

        self.index_t = "uint16"
        self.float_t = "float32"

        self.gltf = pygltflib.GLTF2(
            scene=0,
            scenes=[pygltflib.Scene(nodes=[])],
            nodes=[],
            meshes=[],
            accessors=[],
            materials=[
                pygltflib.Material(
                    name="black",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[0, 0, 0, 1]
                    )
                ),
                pygltflib.Material(
                    name="white",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[1, 1, 1, 1]
                    )
                ),
                pygltflib.Material(
                    name="red",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[1, 0, 0, 1]
                    )
                ),
                pygltflib.Material(
                    name="green",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[0, 1, 0, 1]
                    )
                ),
                pygltflib.Material(
                    name="blue",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[0, 0, 1, 1]
                    )
                ),
                pygltflib.Material(
                    name="gray",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[0.95, 0.95, 0.95, 1]
                    )
                ),
                pygltflib.Material(
                    name="hidden",
                    alphaMode=pygltflib.BLEND,  # Enables transparency
                    doubleSided=True,
#                   extensions={"KHR_materials_unlit": {}},
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[1, 1, 1, 0]
                    )
                )
                #                 pygltflib.Material(
                #                     name="metal",
                #                     doubleSided=True,
                # #                   alphaMode=pygltflib.MASK,
                #                     occlusionTexture=pygltflib.OcclusionTextureInfo(index=1),
                # #                   emissiveFactor=[0.8,0.8,0.8],
                #                     pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                #                         metallicFactor=0.0,
                #                         roughnessFactor=1.0,
                # #                       baseColorFactor=[0.8, 0.8, 0.8, 1],
                #                         baseColorTexture=pygltflib.TextureInfo(index=0),
                #                         metallicRoughnessTexture=pygltflib.TextureInfo(index=1),
                #                     )
                #                 ),
            ],
            bufferViews=[],
            buffers=[pygltflib.Buffer(byteLength=0)],
        )
        self.gltf._glb_data = bytes()


        # Map pairs of (color, alpha) to material's index in material list
        self._color = {(m.name,m.pbrMetallicRoughness.baseColorFactor[3]) if m.pbrMetallicRoughness else (m.name, 0): i
                       for i,m in enumerate(self.gltf.materials)}

        #
        # load assets for steel material
        #
        if False:
            for i,file in enumerate(("rust.jpg", "occlusionRoughnessMetallic.png")):
                path  = str(veux.assets/"metal"/file)
                image = pygltflib.Image()
                image.uri = path
                self.gltf.images.append(image)
                self.gltf.textures.append(pygltflib.Texture(source=i, name=path))

            self.gltf.convert_images(pygltflib.ImageFormat.DATAURI)

    def _init_fixed_node(self, points, triangles, x=True, y=True, z=True):
        # Reuse existing POSITION accessor indices assumed to be stored in self
        points_access = self.position_accessor_index  # previously stored index for POSITION accessor
        indices_access = self.indices_accessor_index  # previously stored index for indices accessor


        # Initialize an array for vertex colors, defaulting to a base color (e.g., gray)
        vertex_colors = np.tile(np.array([0.5, 0.5, 0.5, 1.0], dtype=self.float_t), (len(points), 1))

        # For each triangle, check if the face lies on a plane fixed along given axes
        # and set the corresponding vertices to white if so.
        for tri in triangles:
            # Extract vertices of the triangle
            v0, v1, v2 = points[tri[0]], points[tri[1]], points[tri[2]]

            # Check if all vertices share the same coordinate on any fixed axis:
            if x and (np.isclose(v0[0], v1[0]) and np.isclose(v1[0], v2[0])):
                # Set vertices of this triangle to white
                for vi in tri:
                    vertex_colors[vi] = [1.0, 1.0, 1.0, 1.0]
            if y and (np.isclose(v0[1], v1[1]) and np.isclose(v1[1], v2[1])):
                for vi in tri:
                    vertex_colors[vi] = [1.0, 1.0, 1.0, 1.0]
            if z and (np.isclose(v0[2], v1[2]) and np.isclose(v1[2], v2[2])):
                for vi in tri:
                    vertex_colors[vi] = [1.0, 1.0, 1.0, 1.0]


        # Push the color data to the buffer and create a new accessor for vertex colors
        color_accessor_index = len(self.gltf.accessors)
        self.gltf.accessors.append(
            pygltflib.Accessor(
                bufferView=self._push_data(vertex_colors.astype(self.float_t).tobytes(), pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(vertex_colors),
                type=pygltflib.VEC4,
                max=vertex_colors.max(axis=0).tolist(),
                min=vertex_colors.min(axis=0).tolist(),
            )
        )

        # Create a new material that uses vertex colors directly, e.g., an unlit material.
        # The specifics of creating a material that respects vertex colors depend on your material setup.
        fixed_material = pygltflib.Material(
            name="FixedAxesMaterial",
            pbrMetallicRoughness=pygltflib.PBRMetallicRoughness(
                baseColorFactor=[1.0, 1.0, 1.0, 1.0],  # base color white
                metallicFactor=0.0,
                roughnessFactor=1.0,
            ),
            # Depending on your renderer, you might need to enable vertex colors
        )

        self.gltf.materials.append(fixed_material)
        material_index = len(self.gltf.materials) - 1

        # Create a new mesh primitive using the POSITION and COLOR_0 attributes.
        self.gltf.meshes.append(
            pygltflib.Mesh(
                primitives=[
                    pygltflib.Primitive(
                        mode=pygltflib.TRIANGLES,
                        attributes=pygltflib.Attributes(
                            POSITION=points_access,
                            COLOR_0=color_accessor_index
                        ),
                        material=material_index,
                        indices=indices_access
                    )
                ]
            )
        )
        return len(self.gltf.meshes) - 1


    def _init_nodes(self, style: NodeStyle):
        #
        #
        #
        index_t = self.index_t # "uint8"
        points = style.scale*np.array(
            [
                [-1.0, -1.0,  1.0],
                [ 1.0, -1.0,  1.0],
                [-1.0,  1.0,  1.0],
                [ 1.0,  1.0,  1.0],
                [ 1.0, -1.0, -1.0],
                [-1.0, -1.0, -1.0],
                [ 1.0,  1.0, -1.0],
                [-1.0,  1.0, -1.0],
            ],
            dtype=self.float_t,
        )/10

        triangles = np.array(
            [
                [0, 1, 2],
                [3, 2, 1],
                [1, 0, 4],
                [5, 4, 0],
                [3, 1, 6],
                [4, 6, 1],
                [2, 3, 7],
                [6, 7, 3],
                [0, 2, 5],
                [7, 5, 2],
                [5, 7, 4],
                [6, 4, 7],
            ],
            dtype=index_t,
        )
        triangles_binary_blob = triangles.flatten().tobytes()

        self.gltf.accessors.extend([
            pygltflib.Accessor(
                bufferView=self._push_data(triangles_binary_blob,
                                           pygltflib.ELEMENT_ARRAY_BUFFER),
                componentType=GLTF_T[index_t],
                count=triangles.size,
                type=pygltflib.SCALAR,
                max=[int(triangles.max())],
                min=[int(triangles.min())],
            ),
            pygltflib.Accessor(
                bufferView=self._push_data(points.tobytes(),
                                           pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(points),
                type=pygltflib.VEC3,
                max=points.max(axis=0).tolist(),
                min=points.min(axis=0).tolist(),
            )
        ])


        indices_access = len(self.gltf.accessors)-2 # indices
        points_access  = len(self.gltf.accessors)-1 # points
        self.gltf.meshes.append(
               pygltflib.Mesh(
                 primitives=[
                     pygltflib.Primitive(
                         mode=pygltflib.TRIANGLES,
                         attributes=pygltflib.Attributes(POSITION=points_access),
                         material=self._get_material(style),
                         indices=indices_access
                     )
                 ]
               )
        )

        self._node_mesh = len(self.gltf.meshes) - 1

    def _use_asset(self, name, scale, rotation, material):
        pass

    def _get_material(self, style: DrawStyle)->int:
        if hasattr(style,"alpha"):
            alpha = style.alpha
        else:
            alpha = 1.0

        color = style.color

        if (color, alpha) in self._color:
            return self._color[(color,alpha)]

        elif isinstance(color, str) and color[0] == "#":
            # Remove leading hash in hex
            hx  = color.lstrip("#")
            # Convert hex to RGB
            rgb = [int(hx[i:i+2], 16)/255 for i in (0, 2, 4)]

        elif isinstance(color, tuple):
            rgb = color

        else:
            raise TypeError("Unexpected type for color")

        # Store index for new material
        self._color[(color, alpha)] = len(self.gltf.materials)
        # Create and add new material
        self.gltf.materials.append(
            pygltflib.Material(
                name=str(color),
                doubleSided=True,
                alphaMode=pygltflib.MASK,
                pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                    baseColorFactor=[*rgb, alpha]
                )
            ),
        )
        return self._color[(color, alpha)]


    def _push_data(self, data, target=None, byteStride=None)->int:
        self.gltf.bufferViews.append(
                pygltflib.BufferView(
                    buffer=0,
                    byteStride=byteStride,
                    byteOffset=self.gltf.buffers[0].byteLength,
                    byteLength=len(data),
                    target=target,
                )
        )

        self.gltf._glb_data += data
        self.gltf.buffers[0].byteLength += len(data)
        return len(self.gltf.bufferViews)-1


    def plot_nodes(self, vertices, label = None, style=None, data=None, rotations=None, skin=False, **kwds):
        nodes = []

        if not hasattr(self, "_node_mesh"):
            self._init_nodes(style or NodeStyle())

        if rotations is None:
            rotations = itertools.repeat([0, 0, 0, 1.0])
        else:
            try:
                rotations = Rotation.from_matrix([self._rotation_matrix@R for R in rotations]).as_quat().tolist()
            except ValueError:
                rotations = Rotation.from_rotvec([self._rotation_matrix@R for R in rotations]).as_quat().tolist()


        for coord, rotation in zip(vertices, rotations):
            index = _append_index(self.gltf.nodes, pygltflib.Node(
                    mesh=self._node_mesh,
                    rotation=rotation,
                    translation=(self._rotation_matrix@coord).tolist(),
                )
            )
            if not skin:
                self.gltf.scenes[0].nodes.append(index)
            nodes.append(Node(id=index))

        return nodes

    # def add_joints(self, )

    def add_lines(self, lines: list, style=None, skin_nodes=None):
        """
        Add skinned lines to the glTF object that connect pairs of nodes specified in `lines`.
        The lines will deform as the corresponding nodes are translated.

        :param gltf: The pygltflib.GLTF2 object to modify.
        :param lines: A list of pairs of indices (i, j), where i and j are node indices.
        :param access_vertices: The accessor index for initial positions of the nodes.
        """
        gltf = self.gltf
        scene = gltf.scenes[0]
        EYE4 = np.eye(4, dtype=self.float_t)

        # Validate that all node indices in `lines` exist in gltf.nodes
        max_node_idx = len(gltf.nodes) - 1
        for i, j in lines:
            if i > max_node_idx or j > max_node_idx:
                raise ValueError(f"Node indices {i} or {j} in `lines` are out of range for `gltf.nodes`.")


        # Create joints (one per node) and bind the lines to them
        if skin_nodes is not None:
            joint_nodes = list({skin_nodes[idx].id for pair in lines for idx in pair})  # Unique joint indices
        else:
            joint_nodes = list({idx for pair in lines for idx in pair})  # Unique joint indices

        joint_node_to_index = {node: i for i, node in enumerate(joint_nodes)}

        points = np.zeros((len(joint_nodes), 3), dtype=self.float_t)
        # points = np.zeros((1,3), dtype=self.float_t) 
        # points = np.random.random((len(joint_nodes), 3))
        # for i in range(len(joint_nodes)):
        #     points[i] /= np.linalg.norm(points[i])*1e3
        # points = points.astype(self.float_t)

        # points = np.array([self._rotation_matrix.T@self.gltf.nodes[n].translation for n in joint_nodes], dtype=self.float_t)
        # points = np.array(_,dtype=self.float_t) #np.array([self._rotation_matrix.T@n for n in _], dtype=self.float_t)
        ver_accessor = _append_index(self.gltf.accessors, pygltflib.Accessor(
                bufferView=self._push_data(points.tobytes(), pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(points),
                type=pygltflib.VEC3,
                max=points.max(axis=0).tolist(),
                min=points.min(axis=0).tolist(),
            )
        )
    
        # Create a root node for the entire skeleton
        skeleton_root_node = pygltflib.Node(name="LineSkeleton", 
                                            children=joint_nodes)
        skeleton_root_idx = _append_index(gltf.nodes, skeleton_root_node)
        scene.nodes.append(skeleton_root_idx)

        # Create the inverse bind matrices
        inverse_bind_matrices = []
        for joint_node in joint_nodes:
            # Compute the transform matrix for the bind pose
            node = gltf.nodes[joint_node]
            t_matrix = np.eye(4, dtype=self.float_t)
            t_matrix[:3, 3] = node.translation if node.translation else [0.0, 0.0, 0.0]
            rotation_matrix = Rotation.from_quat(node.rotation).as_matrix() if node.rotation else EYE3
            t_matrix[:3, :3] = rotation_matrix #* (node.scale if node.scale else 1.0)
            # t_matrix = np.linalg.inv(t_matrix)

            # inverse_bind_matrices.append(np.array(t_matrix, dtype=self.float_t).T)
            inverse_bind_matrices.append(EYE4)

        # Add buffer view and accessor for the inverse bind matrices
        # Flatten inverse bind matrices for glTF format
        ibm_array = np.array(inverse_bind_matrices, dtype=self.float_t).reshape(-1, 16)
        # ibm_array = np.array([ibm.flatten() for ibm in inverse_bind_matrices], dtype=self.float_t)

        ibm_accessor_idx = _append_index(gltf.accessors, pygltflib.Accessor(
            bufferView=self._push_data(ibm_array.tobytes(), target=None),
            componentType=GLTF_T[self.float_t],
            count=len(joint_nodes),
            type="MAT4"
        ))

        # Create the skin
        if not gltf.skins:
            gltf.skins = []
        skin_idx = len(gltf.skins)
        gltf.skins.append(pygltflib.Skin(
            inverseBindMatrices=ibm_accessor_idx,
            joints=joint_nodes,
            skeleton=skeleton_root_idx,  # Root joint
            name="LineSkin"
        ))

        # Create the index buffer for the lines
        # indices = []
        joints_0 = []
        weights_0 = []
        for i, j in lines:
            # indices.extend([i, j])
            joints_0.extend([
                [joint_node_to_index[i], 0, 0, 0],
                [joint_node_to_index[j], 0, 0, 0]
            ])
            weights_0.extend([
                [1.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0]
            ])

        # index_array = np.array(indices, dtype=self.index_t)#.flatten()
        index_array = np.array(lines, dtype=self.index_t).reshape(-1)#.flatten()
        # index_array = np.zeros((len(lines),2), dtype=self.index_t).reshape(-1)

        # Add buffer view and accessor for the line indices

        index_accessor_idx = len(gltf.accessors)
        gltf.accessors.append(pygltflib.Accessor(
            bufferView=self._push_data(index_array.tobytes(),pygltflib.ELEMENT_ARRAY_BUFFER),
            componentType=GLTF_T[self.index_t],
            count=len(index_array),
            type="SCALAR",
            min=[int(index_array.min())],
            max=[int(index_array.max())]
        ))

        # Create the line mesh with skinning attributes
        joints_0_array  = np.array(joints_0, dtype=self.index_t)

        joints_0_accessor_idx = len(gltf.accessors)
        gltf.accessors.append(pygltflib.Accessor(
            bufferView=self._push_data(joints_0_array.tobytes(), pygltflib.ARRAY_BUFFER),
            componentType=GLTF_T[self.index_t],
            count=len(joints_0),
            type="VEC4"
        ))

        weights_0_array = np.array(weights_0, dtype=self.float_t)
        weights_0_accessor_idx = len(gltf.accessors)
        gltf.accessors.append(pygltflib.Accessor(
            bufferView=self._push_data(weights_0_array.tobytes(), pygltflib.ARRAY_BUFFER),
            componentType=GLTF_T[self.float_t],
            count=len(weights_0),
            type="VEC4"
        ))

        # Create the line mesh
        mesh_idx = len(gltf.meshes)
        line_mesh = pygltflib.Mesh(
            primitives=[
                pygltflib.Primitive(
                    attributes=pygltflib.Attributes(
                        POSITION=ver_accessor,
                        JOINTS_0=joints_0_accessor_idx,
                        WEIGHTS_0=weights_0_accessor_idx
                    ),
                    indices=index_accessor_idx,
                    mode=pygltflib.LINES,
                    material=self._get_material(style or LineStyle())
                )
            ], name="LineSkinMesh")
        gltf.meshes.append(line_mesh)

        # Create a node for the line mesh
        mesh_node = _append_index(gltf.nodes, pygltflib.Node(
            mesh=mesh_idx, 
            skin=skin_idx, 
            # rotation=self._rotation,
            name="LineSkinMeshNode"
        ))

        # Add the new node to the scene
        scene.nodes.append(mesh_node)


    def plot_lines(self, vertices, indices=None, style: LineStyle=None, vcache:str=None, **kwds):

        lines = []
        material = self._get_material(style or LineStyle())

        # vertices is given with nans separating line groups, but
        # GLTF does not accept nans so we have to filter these
        # out, and add distinct meshes for each line group
        assert np.all(np.isnan(vertices[np.isnan(vertices[:,0]), :]))
        points  = np.array(vertices[~np.isnan(vertices[:,0]),:], dtype=self.float_t)

        if points.size == 0:
            return

        self.gltf.accessors.append(
            pygltflib.Accessor(
                bufferView=self._push_data(points.tobytes(), pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(points),
                type=pygltflib.VEC3,
                max=points.max(axis=0).tolist(),
                min=points.min(axis=0).tolist(),
            )
        )
        points_access = len(self.gltf.accessors) - 1

        if indices is None:
            # Get indices by splitting at nans
            indices_ = utility.split(np.arange(len(vertices), dtype=self.index_t), np.nan, vertices[:,0])
        else:
            indices_ = list(map(lambda x: np.array(x, dtype=self.index_t), indices))

        for indices in indices_:
            # Here, n adjusts indices by the number of nan rows that were removed so far
            n  = sum(np.isnan(vertices[:indices[0],0]))
            indices_array = indices - np.dtype(self.index_t).type(n)

            indices_binary_blob = indices_array.tobytes()

            if len(indices_array) <= 1:
                import warnings
                warnings.warn(indices_array, file=sys.stderr)
                continue

            self.gltf.accessors.extend([
                pygltflib.Accessor(
                    bufferView=self._push_data(indices_binary_blob,
                                               pygltflib.ELEMENT_ARRAY_BUFFER),
                    componentType=GLTF_T[self.index_t],
                    count=indices_array.size,
                    type=pygltflib.SCALAR,
                    max=[int(indices_array.max())],
                    min=[int(indices_array.min())],
                )
            ])
            self.gltf.meshes.append(
                   pygltflib.Mesh(
                     primitives=[
                         pygltflib.Primitive(
                             mode=pygltflib.LINE_STRIP,
                             attributes=pygltflib.Attributes(POSITION=points_access),
                             material=material,
                             # most recently added accessor
                             indices=len(self.gltf.accessors)-1,
                             # TODO: use this mechanism to add annotation data
                             extras={},
                         )
                     ]
                   )
            )

            self.gltf.nodes.append(pygltflib.Node(
                    mesh=len(self.gltf.meshes)-1,
                    rotation=self._rotation
                )
            )

            self.gltf.scenes[0].nodes.append(
                len(self.gltf.nodes)-1
            )

            lines.append(Line(
                id=len(self.gltf.nodes)-1
            ))

        return lines
    
    def draw_skin(self, vertices, triangles):
        pass

    def plot_mesh(self, vertices, triangles, local_coords=None, style=None, **kwds) -> tuple:

        material  = self._get_material(style or MeshStyle())

        if isinstance(triangles, int):
            index_access = triangles
        else:
            triangles = np.array(triangles,dtype=self.index_t)
            self.gltf.accessors.extend([
                pygltflib.Accessor(
                    bufferView=self._push_data(triangles.flatten().tobytes(), pygltflib.ELEMENT_ARRAY_BUFFER),
                    componentType=GLTF_T[self.index_t],
                    count=triangles.size,
                    type=pygltflib.SCALAR,
                    max=[int(triangles.max())],
                    min=[int(triangles.min())],
                )
            ])
            index_access = len(self.gltf.accessors)-1

        if isinstance(vertices, int):
            point_access = vertices
        else:
            points    = np.array(vertices, dtype=self.float_t)
            self.gltf.accessors.extend([
                pygltflib.Accessor(
                    bufferView=self._push_data(points.tobytes(), pygltflib.ARRAY_BUFFER),
                    componentType=GLTF_T[self.float_t],
                    count=len(points),
                    type=pygltflib.VEC3,
                    max=points.max(axis=0).tolist(),
                    min=points.min(axis=0).tolist(),
                )
            ])
            point_access = len(self.gltf.accessors)-1

        # Add accessors for (1) point coordinates and (2) indices

        self.gltf.meshes.append(
               pygltflib.Mesh(
                 primitives=[
                     pygltflib.Primitive(
                         mode=pygltflib.TRIANGLES,
                         attributes=pygltflib.Attributes(POSITION=point_access),
                         material=material,
                         indices=index_access,
                         targets=[
                             # TODO: implement morph targets
#                           pygltflib.Attributes(POSITION=point_access)
                         ]
                     )
                 ]
               )
        )

        if local_coords is not None:
            locoor = np.array(local_coords, dtype=self.float_t)
            self.gltf.accessors.extend([
                pygltflib.Accessor(
                    bufferView=self._push_data(locoor.tobytes(), pygltflib.ARRAY_BUFFER),
                    componentType=GLTF_T[self.float_t],
                    count=len(locoor),
                    type=pygltflib.VEC2,
                    max=locoor.max(axis=0).tolist(),
                    min=locoor.min(axis=0).tolist(),
                )
            ])
            self.gltf.meshes[-1].primitives[0].attributes.TEXCOORD_0 = len(self.gltf.accessors) -1

        self.gltf.nodes.append(pygltflib.Node(
                mesh=len(self.gltf.meshes)-1,
                rotation=self._rotation
            )
        )

        scene_node = len(self.gltf.nodes)-1
        self.gltf.scenes[0].nodes.append(scene_node)

        return Mesh(id=scene_node,
                    vertices=point_access, 
                    indices=index_access)


    def plot_mesh_field(self, mesh_handle, field,
                        colormap="viridis",
                        vmin=None, vmax=None,
                        **kwds) -> tuple:
        """
        Draw a mesh colored by a scalar field.
        
        :param vertices: Nx3 array of vertex positions
        :param triangles: Mx3 array of triangle indices
                          (or an int referencing an existing accessor)
        :param field: length-N array (one scalar per vertex)
        :param style: optional style info
        :param colormap: name of a matplotlib colormap or something similar
        :param vmin, vmax: optionally set the data range for mapping
        :param kwds: other keywords
        """

        # Step 1. Determine color array from 'field'
        #         For demonstration, let's assume you have a helper
        #         that returns Nx4 RGBA in [0,1].
        color_array = self._map_field_to_colors(field, colormap, vmin, vmax)

        # Step 2. Plot the mesh (re-use plot_mesh) but we'll also attach colors
        # mesh_handle = self.plot_mesh(vertices, triangles, style=style, **kwds)

        # Step 3. Add color data as a separate accessor in the just-created mesh
        color_array = color_array.astype(self.float_t)
        self.gltf.accessors.extend([
            pygltflib.Accessor(
                bufferView=self._push_data(color_array.tobytes(), pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(color_array),
                type=pygltflib.VEC4,  # RGBA
                max=color_array.max(axis=0).tolist(),
                min=color_array.min(axis=0).tolist(),
            )
        ])
        color_access = len(self.gltf.accessors) - 1

        # Attach the color accessor to the existing mesh’s attributes
        mesh_idx = self.gltf.nodes[mesh_handle.id].mesh
        self.gltf.meshes[mesh_idx].primitives[0].attributes.COLOR_0 = color_access

        return mesh_handle


    def _map_field_to_colors(self, field, colormap="viridis", vmin=None, vmax=None):
        """
        Convert 1D array of scalar values into Nx4 RGBA array in [0,1].
        For example, you might rely on matplotlib here, or any
        color-mapping library. For demonstration, let's assume we have:

        >>> import matplotlib.pyplot as plt
        >>> cmap = plt.get_cmap(colormap)
        >>> # we then map normalized values to RGBA
        """
        import matplotlib
        import matplotlib.cm

        field = np.array(field, dtype=np.float64)
        if vmin is None:
            vmin = field.min()
        if vmax is None:
            vmax = field.max()

        norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax, clip=True)
        cmap = matplotlib.cm.get_cmap(colormap)
        
        # shape: (N, 4)
        colors = cmap(norm(field))
        # ensure Nx4
        return colors


    def to_glb(self)->bytes:
        return b"".join(self.gltf.save_to_bytes())

    def write(self, filename=None):

        self.gltf.save(filename)

#       if "glb" in filename[-3:]:
#           glb = b"".join(self.gltf.save_to_bytes())
#           with open(filename,"wb+") as f:
#               f.write(glb)




# Claudio Perez
# Summer 2024
import base64
import textwrap
from pathlib import Path

class Viewer:
    def __init__(self, viewer=None, path=None, data=None):
        if data is not None:
            data64 = base64.b64encode(data).decode('utf-8')
            self._glbsrc=f"data:model/gltf-binary;base64,{data64}"
        else:
            self._glbsrc = path 

        self._viewer = viewer if viewer is not None else "mv"

    def get_html(self):
        if self._viewer == "babylon":
            with open(Path(__file__).parents[0]/"babylon.html", "r") as f:
                return f.read()

        if self._viewer == "three-170":
            with open(Path(__file__).parents[0]/"three-170.html", "r") as f:
                return f.read()

        if self._viewer == "three-160":
            with open(Path(__file__).parents[0]/"gltf.html", "r") as f:
                return f.read()

        elif self._viewer == "three-130":
            with open(Path(__file__).parents[0]/"index.html", "r") as f:
                return f.read()

        elif self._viewer == "mv":
            return _model_viewer(self._glbsrc, control=False)
            html = textwrap.dedent("""
          <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
            "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
          <html xmlns="http://www.w3.org/1999/xhtml" lang="en">
          <head>
            <meta charset="utf-8">
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <title>veux</title>
            <script type="module" src="https://ajax.googleapis.com/ajax/libs/model-viewer/4.0.0/model-viewer.min.js"></script>
            """ + """
            <style>
              .controls {
                display: flex;
                justify-content: center;
                gap: 10px;
                margin-top: 10px;
              }

              .controls button {
                background-color: #4CAF50;
                color: white;
                border: none;
                border-radius: 5px;
                padding: 10px 20px;
                font-size: 16px;
                cursor: pointer;
                transition: background-color 0.3s;
              }

              .controls button:hover {
                background-color: #45a049;
              }

              .controls button:disabled {
                background-color: #cccccc;
                cursor: not-allowed;
              }
            </style>
          </head>
          <body>
          """ + f"""
            <model-viewer id="veux-viewer"
                          alt="rendering"
                          src="{self._glbsrc}"
                          autoplay
                          style="width: 100%; height: 500px;"
                          max-pixel-ratio="2"
                          shadow-intensity="1"
                          environment-image="/black_ground.hdr"
                          environment-image="neutral"
                          shadow-light="10000 10000 10000"
                          exposure="0.8"
                          camera-controls
                          min-camera-orbit="auto auto 0m"
                          touch-action="pan-y">
            </model-viewer>
        <!--
                          bounds="0 0 0 10000 10000 10000"
                          ar
                          environment-image="https://modelviewer.dev/shared-assets/environments/spruit_sunrise_1k_LDR.jpg"
                          shadow-softness="0.8"
                          ar-modes="scene-viewer; quick-look"
                          position="0 0 -50"
                          scale="0.01 0.01 0.01"
        -->
            """ + """

              <div class="controls">
                <button id="step-backward">Step Left</button>
                <button id="toggle-animation">Pause</button>
                <button id="step-forward">Step Right</button>
              </div>

            <script>
              const modelViewer = document.getElementById('veux-viewer');
              const toggleButton = document.getElementById('toggle-animation');
              const stepBackwardButton = document.getElementById('step-backward');
              const stepForwardButton = document.getElementById('step-forward');

              toggleButton.addEventListener('click', () => {
                if (modelViewer.paused) {
                  modelViewer.play();
                  toggleButton.textContent = 'Pause';
                } else {
                  modelViewer.pause();
                  toggleButton.textContent = 'Play';
                }
              });

              stepBackwardButton.addEventListener('click', () => {
                if (modelViewer.currentTime > 0) {
                  modelViewer.currentTime -= 0.1; // Step backward by 0.1 seconds
                }
              });

              stepForwardButton.addEventListener('click', () => {
                if (modelViewer.currentTime < modelViewer.totalTime) {
                  modelViewer.currentTime += 0.1; // Step forward by 0.1 seconds
                }
              });

              // Disable step buttons when out of range
              modelViewer.addEventListener('timeupdate', () => {
                stepBackwardButton.disabled = modelViewer.currentTime <= 0;
                stepForwardButton.disabled = modelViewer.currentTime >= modelViewer.totalTime;
              });
            </script>
          </body>
          </html>
        """)
        return html

def _model_viewer(source, control=False):
      library = '<script type="module" src="https://ajax.googleapis.com/ajax/libs/model-viewer/4.0.0/model-viewer.min.js"></script>'

      with open(Path(__file__).parents[0]/"controls.css", "r") as f:
          control_style = f"<style>{f.read()}</style>"

      with open(Path(__file__).parents[0]/"controls.js", "r") as f:
          control_code = f"<script>{f.read()}</script>"

      control_html = """
        <div class="controls">
          <!-- <button id="step-backward">Step Left</button> -->
          <button id="toggle-animation">Pause</button>
          <!-- <button id="step-forward">Step Right</button> -->
        </div>
      """


      foot = "</body></html>"
      head = f"""
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
          "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
        <html xmlns="http://www.w3.org/1999/xhtml" lang="en">
        <head>
          <meta charset="utf-8">
          <meta name="viewport" content="width=device-width, initial-scale=1">
          <title>veux</title>
          {library}
          {control_style}
        </head>
        <body>
      """

      viewer = f"""
          <model-viewer id="veux-viewer"
                        alt="rendering"
                        src="{source}"
                        autoplay
                        style="width: 100%; height: 500px;"
                        max-pixel-ratio="2"
                        shadow-intensity="1"
                        environment-image="/black_ground.hdr"
                        environment-image="neutral"
                        shadow-light="10000 10000 10000"
                        exposure="0.8"
                        camera-controls
                        min-camera-orbit="auto auto 0m"
                        touch-action="pan-y">
          </model-viewer>
      """
      return textwrap.dedent(f"""
            {head}
            {viewer}
            {control_html if control else ""}
            {control_code if control else ""}
            {foot}
      """)



"""
A direct port of mapbox/earcut

Adapted from:
  https://github.com/joshuaskelly/earcut-python
"""
import math
import numpy as np

__all__ = ['earcut', 'deviation', 'flatten']

def earcut1(vertices):
    """
    Triangulate a simple polygon using the mapbox_earcut library.

    Parameters:
      vertices (list or ndarray): A list or array of [x, y] coordinates of the polygon vertices in order.

    Returns:
      triangles (ndarray): An (n, 3) array of indices into `vertices`, each row representing a triangle.
    """
    import mapbox_earcut as earcut
    if np.array_equal(vertices[0], vertices[-1]):
        vertices = vertices[:-1]

    # Convert the list of coordinates to a flat list
    flattened = np.array(vertices).reshape(-1,2) #.flatten()

    # Perform triangulation
    triangle_indices = earcut.triangulate_float64(flattened, np.array([len(vertices)]))

    # Convert the flat list of indices into a (n, 3) array
    triangles = np.array(triangle_indices).reshape(-1, 3)

    return triangles

class _PointLink:
    def __init__(self, i, x, y):
    # vertice index in coordinates array
        self.i = i

        # vertex coordinates
        self.x = x
        self.y = y

        # previous and next vertice nodes in a polygon ring
        self.prev = None
        self.next = None

        # z-order curve value
        self.z = None

        # previous and next nodes in z-order
        self.prevZ = None
        self.nextZ = None

        # indicates whether this is a steiner point
        self.steiner = False
    
    def append(self, data):
        pass

    def __getitem__(self, i):
        if i == 0:
            return self.x 
        if i == 1: 
            return self.y 
        else:
            raise IndexError

    @property 
    def shape(self):
        # TODO: Return coordinates
        return


def _point_in_triangle(a, b, c, p):
    "check if a point lies within a convex triangle"
    ax, ay = a 
    bx, by = b
    cx, cy = c
    px, py = p
    return (cx - px) * (ay - py) - (ax - px) * (cy - py) >= 0 and \
           (ax - px) * (by - py) - (bx - px) * (ay - py) >= 0 and \
           (bx - px) * (cy - py) - (cx - px) * (by - py) >= 0


def _valid_diagonal(a, b):
    "check if a diagonal between two polygon nodes is valid (lies in polygon interior)"
    return a.next.i != b.i and a.prev.i != b.i and not intersectsPolygon(a, b) and \
        locallyInside(a, b) and locallyInside(b, a) and middleInside(a, b)


def _points_equal(a, b):
    "check if two points are equal"
    return a[0] == b[0] and a[1] == b[1]



def intersects(p1, q1, p2, q2):
    "check if two segments intersect"
    if (_points_equal(p1, q1) and _points_equal(p2, q2)) or (_points_equal(p1, q2) and _points_equal(p2, q1)):
        return True

    return _triangle_area(p1, q1, p2) > 0 != _triangle_area(p1, q1, q2) > 0 and \
           _triangle_area(p2, q2, p1) > 0 != _triangle_area(p2, q2, q1) > 0

# def compareX(a, b):
#     return a.x - b.x

def intersectsPolygon(a, b):
    "check if a polygon diagonal intersects any polygon segments"
    do = True
    p = a

    while do or p != a:
        do = False
        if (p.i != a.i and p.next.i != a.i and p.i != b.i and p.next.i != b.i and intersects(p, p.next, a, b)):
            return True

        p = p.next

    return False


def locallyInside(a, b):
    "check if a polygon diagonal is locally inside the polygon"
    if _triangle_area(a.prev, a, a.next) < 0:
        return  _triangle_area(a, b, a.next) >= 0 and _triangle_area(a, a.prev, b) >= 0
    else:
        return _triangle_area(a, b, a.prev) < 0 or _triangle_area(a, a.next, b) < 0


def middleInside(a, b):
    "check if the middle point of a polygon diagonal is inside the polygon"
    do = True
    p = a
    inside = False
    px = (a.x + b.x) / 2
    py = (a.y + b.y) / 2

    while do or p != a:
        do = False
        if ((p.y > py) != (p.next.y > py)) and (px < (p.next.x - p.x) * (py - p.y) / (p.next.y - p.y) + p.x):
            inside = not inside

        p = p.next

    return inside

# link two polygon vertices with a bridge; if the vertices belong to the same ring, it splits polygon into two;
# if one belongs to the outer ring and another to a hole, it merges it into a single ring
def _split_polygon(a, b):
    a2 = _PointLink(a.i, a.x, a.y)
    b2 = _PointLink(b.i, b.x, b.y)
    an = a.next
    bp = b.prev

    a.next = b
    b.prev = a

    a2.next = an
    an.prev = a2

    b2.next = a2
    a2.prev = b2

    bp.next = b2
    b2.prev = bp

    return b2

def earcut(vertices, holeIndices=None, dim=None):
    # Convert to numpy array
    vertices = np.asarray(vertices)

    # Remove duplicate last point if it's the same as the first
    if np.array_equal(vertices[0], vertices[-1]):
        vertices = vertices[:-1]

    dim = dim or 2

    data = vertices.flatten() #reshape(-1,2)

    hasHoles = holeIndices and len(holeIndices)
    outerLen =  holeIndices[0] * dim if hasHoles else len(data)
    outerNode = _create_links(data, 0, outerLen, dim, True)
    triangles = []

    if not outerNode:
        return triangles

    minX = None
    minY = None
    maxX = None
    maxY = None
    x = None
    y = None
    size = None

    if hasHoles:
        outerNode = _purge_holes(data, holeIndices, outerNode, dim)

    # if the shape is not too simple, we'll use z-order curve hash later; calculate polygon bbox
    if len(data) > 80 * dim:
        minX = maxX = data[0]
        minY = maxY = data[1]

        for i in range(dim, outerLen, dim):
            x = data[i]
            y = data[i + 1]
            if x < minX:
                minX = x
            if y < minY:
                minY = y
            if x > maxX:
                maxX = x
            if y > maxY:
                maxY = y

        # minX, minY and size are later used to transform coords into integers for z-order calculation
        size = max(maxX - minX, maxY - minY)

    _earcut_links(outerNode, triangles, dim, minX, minY, size)

    return triangles


def _create_links(data, start, end, dim, clockwise):
    "create a circular doubly linked list from polygon points in the specified winding order"
    i = None
    last = None

    if (clockwise == (_polygon_area(data, start, end, dim) > 0)):
        for i in range(start, end, dim):
            last = _insert_link(i, data[i], data[i + 1], last)

    else:
        for i in reversed(range(start, end, dim)):
            last = _insert_link(i, data[i], data[i + 1], last)

    if last and _points_equal(last, last.next):
        _remove_link(last)
        last = last.next

    return last


def _filter_points(start, end=None):
    "eliminate colinear or duplicate points"
    if not start:
        return start
    if not end:
        end = start

    p = start
    again = True

    while again or p != end:
        again = False

        if not p.steiner and (_points_equal(p, p.next) or _triangle_area(p.prev, p, p.next) == 0):
            _remove_link(p)
            p = end = p.prev
            if (p == p.next):
                return None

            again = True

        else:
            p = p.next

    return end

def _earcut_links(ear: "_PointLink", triangles, dim, minX, minY, size, _pass=None):
    """main ear slicing loop which triangulates a polygon (given as a linked list)"""

    if not ear:
        return

    # interlink polygon nodes in z-order
    if not _pass and size:
        _index_curve(ear, minX, minY, size)

    stop = ear
    prev = None
    next = None

    # iterate through ears, slicing them one by one
    while ear.prev != ear.next:
        prev = ear.prev
        next = ear.next

        if isEarHashed(ear, minX, minY, size) if size else _is_ear(ear):
            # cut off the triangle
            triangles.append(prev.i // dim)
            triangles.append(ear.i  // dim)
            triangles.append(next.i // dim)

            _remove_link(ear)

            # skipping the next vertice leads to less sliver triangles
            ear  = next.next
            stop = next.next

            continue

        ear = next

        # if we looped through the whole remaining polygon and can't find any more ears
        if ear == stop:
            # try filtering points and slicing again
            if not _pass:
                _earcut_links(_filter_points(ear), triangles, dim, minX, minY, size, 1)

                # if this didn't work, try curing all small self-intersections locally
            elif _pass == 1:
                ear = cureLocalIntersections(ear, triangles, dim)
                _earcut_links(ear, triangles, dim, minX, minY, size, 2)

                # as a last resort, try splitting the remaining polygon into two
            elif _pass == 2:
                splitEarcut(ear, triangles, dim, minX, minY, size)

            break

# check whether a polygon node forms a valid ear with adjacent nodes
def _is_ear(ear):
    a = ear.prev
    b = ear
    c = ear.next

    if _triangle_area(a, b, c) >= 0:
        return False # reflex, can't be an ear

    # now make sure we don't have other points inside the potential ear
    p = ear.next.next

    while p != ear.prev:
        if _point_in_triangle(a, b, c, p) and _triangle_area(p.prev, p, p.next) >= 0:
                return False
        p = p.next

    return True

def isEarHashed(ear, minX, minY, size):
    a = ear.prev
    b = ear
    c = ear.next

    if _triangle_area(a, b, c) >= 0:
        return False # reflex, can't be an ear

    # triangle bbox; min & max were calculated like this for speed
    minTX = (a[0] if a[0] < c[0] else c[0]) if a[0] < b[0] else (b[0] if b[0] < c[0] else c[0])
    minTY = (a[1] if a[1] < c[1] else c[1]) if a[1] < b[1] else (b[1] if b[1] < c[1] else c[1])
    maxTX = (a[0] if a[0] > c[0] else c[0]) if a[0] > b[0] else (b[0] if b[0] > c[0] else c[0])
    maxTY = (a[1] if a[1] > c[1] else c[1]) if a[1] > b[1] else (b[1] if b[1] > c[1] else c[1])

    # z-order range for the current triangle bbox;
    minZ = _z_order(minTX, minTY, minX, minY, size)
    maxZ = _z_order(maxTX, maxTY, minX, minY, size)

    # first look for points inside the triangle in increasing z-order
    p = ear.nextZ

    while p and p.z <= maxZ:
        if p != ear.prev and p != ear.next and _point_in_triangle(a, b, c, p) and _triangle_area(p.prev, p, p.next) >= 0:
            return False
        p = p.nextZ

    # then look for points in decreasing z-order
    p = ear.prevZ

    while p and p.z >= minZ:
        if p != ear.prev and p != ear.next and _point_in_triangle(a, b, c, p) and _triangle_area(p.prev, p, p.next) >= 0:
            return False
        p = p.prevZ

    return True

# go through all polygon nodes and cure small local self-intersections
def cureLocalIntersections(start, triangles, dim):
    do = True
    p = start

    while do or p != start:
        do = False

        a = p.prev
        b = p.next.next

        if not _points_equal(a, b) and intersects(a, p, p.next, b) and locallyInside(a, b) and locallyInside(b, a):
            triangles.append(a.i // dim)
            triangles.append(p.i // dim)
            triangles.append(b.i // dim)

            # remove two nodes involved
            _remove_link(p)
            _remove_link(p.next)

            p = start = b

        p = p.next

    return p

# try splitting polygon into two and triangulate them independently
def splitEarcut(start, triangles, dim, minX, minY, size):
    # look for a valid diagonal that divides the polygon into two
    do = True
    a = start

    while do or a != start:
        do = False
        b = a.next.next

        while b != a.prev:
            if a.i != b.i and _valid_diagonal(a, b):
                # split the polygon in two by the diagonal
                c = _split_polygon(a, b)

                # filter colinear points around the cuts
                a = _filter_points(a, a.next)
                c = _filter_points(c, c.next)

                # run earcut on each half
                _earcut_links(a, triangles, dim, minX, minY, size)
                _earcut_links(c, triangles, dim, minX, minY, size)
                return

            b = b.next

        a = a.next


def _purge_holes(data, hole_indices, outerNode, dim):
    "Link every hole into the outer loop, producing a single-ring polygon without holes"
    queue = []
    i = None
    hole_count = len(hole_indices)
    start = None
    end = None
    lst = None

    for i in range(hole_count):
        start = hole_indices[i] * dim
        end =  hole_indices[i + 1] * dim if i < hole_count - 1 else len(data)
        lst = _create_links(data, start, end, dim, False)

        if (lst == lst.next):
            lst.steiner = True

        queue.append(_leftmost_node(lst))

    queue = sorted(queue, key=lambda i: i[0])

    # process holes from left to right
    for i in range(len(queue)):
        _remove_hole(queue[i], outerNode)
        outerNode = _filter_points(outerNode, outerNode.next)

    return outerNode


def _remove_hole(hole, outer_node):
    "find a bridge between vertices that connects hole with an outer ring and and link it"
    outer_node = _find_hole_bridge(hole, outer_node)
    if outer_node:
        b = _split_polygon(outer_node, hole)
        _filter_points(b, b.next)


def _find_hole_bridge(hole, outerNode):
    "David Eberly's algorithm for finding a bridge between hole and outer polygon"
    do = True
    p = outerNode
    hx = hole.x
    hy = hole.y
    qx = -math.inf
    m = None

    # find a segment intersected by a ray from the hole's leftmost point to the left;
    # segment's endpoint with lesser x will be potential connection point
    while do or p != outerNode:
        do = False
        if hy <= p.y and hy >= p.next.y and p.next.y - p.y != 0:
            x = p.x + (hy - p.y) * (p.next.x - p.x) / (p.next.y - p.y)

            if x <= hx and x > qx:
                qx = x

                if (x == hx):
                    if hy == p.y:
                        return p
                    if hy == p.next.y:
                        return p.next

                m = p if p.x < p.next.x else p.next

        p = p.next

    if not m:
        return None

    if hx == qx:
        return m.prev # hole touches outer segment; pick lower endpoint

    # look for points inside the triangle of hole point, segment intersection and endpoint;
    # if there are no points found, we have a valid connection;
    # otherwise choose the point of the minimum angle with the ray as connection point

    stop = m
    mx = m.x
    my = m.y
    tanMin = math.inf
    tan = None

    p = m.next

    while p != stop:
        hx_or_qx = hx if hy < my else qx
        qx_or_hx = qx if hy < my else hx

        if hx >= p.x and p.x >= mx and _point_in_triangle((hx_or_qx, hy), m, (qx_or_hx, hy), p):

            tan = abs(hy - p.y) / (hx - p.x) # tangential

            if (tan < tanMin or (tan == tanMin and p.x > m.x)) and locallyInside(p, hole):
                m = p
                tanMin = tan

        p = p.next

    return m

# interlink polygon nodes in z-order
def _index_curve(start, minX, minY, size):
    do = True
    p = start

    while do or p != start:
        do = False

        if p.z == None:
            p.z = _z_order(p.x, p.y, minX, minY, size)

        p.prevZ = p.prev
        p.nextZ = p.next
        p = p.next

    p.prevZ.nextZ = None
    p.prevZ = None

    _sort_links(p)

def _sort_links(L):
    """
    Simon Tatham's linked list merge sort algorithm
    http://www.chiark.greenend.org.uk/~sgtatham/algorithms/listsort.html
    """
    do = True
    i = None
    p = None
    q = None
    e = None
    tail = None
    merge_count = None
    pSize = None
    qSize = None
    inSize = 1

    while do or merge_count > 1:
        do = False
        p = L
        L = None
        tail = None
        merge_count = 0

        while p:
            merge_count += 1
            q = p
            pSize = 0
            for i in range(inSize):
                pSize += 1
                q = q.nextZ
                if not q:
                    break

            qSize = inSize

            while pSize > 0 or qSize > 0 and q:

                if pSize == 0:
                    e = q
                    q = q.nextZ
                    qSize -= 1

                elif qSize == 0 or not q:
                    e = p
                    p = p.nextZ
                    pSize -= 1

                elif p.z <= q.z:
                    e = p
                    p = p.nextZ
                    pSize -= 1

                else:
                    e = q
                    q = q.nextZ
                    qSize -= 1

                if tail:
                    tail.nextZ = e

                else:
                    L = e

                e.prevZ = tail
                tail = e

            p = q

        tail.nextZ = None
        inSize *= 2

    return L


def _z_order(x, y, minX, minY, size):
    "z-order of a point given coords and size of the data bounding box"

    # coords are transformed into non-negative 15-bit integer range
    x = 32767 * (x - minX) // size
    y = 32767 * (y - minY) // size

    x = (x | (x << 8)) & 0x00FF00FF
    x = (x | (x << 4)) & 0x0F0F0F0F
    x = (x | (x << 2)) & 0x33333333
    x = (x | (x << 1)) & 0x55555555

    y = (y | (y << 8)) & 0x00FF00FF
    y = (y | (y << 4)) & 0x0F0F0F0F
    y = (y | (y << 2)) & 0x33333333
    y = (y | (y << 1)) & 0x55555555

    return x | (y << 1)


def _leftmost_node(start):
    "find the leftmost node of a polygon ring"
    do = True
    p = start
    leftmost = start

    while do or p != start:
        do = False
        if p.x < leftmost.x:
            leftmost = p
        p = p.next

    return leftmost


def _insert_link(i, x, y, last):
    "create a node and optionally link it with previous one (in a circular doubly linked list)"
    p = _PointLink(i, x, y)

    if not last:
        p.prev = p
        p.next = p

    else:
        p.next = last.next
        p.prev = last
        last.next.prev = p
        last.next = p

    return p

def _remove_link(p):
    p.next.prev = p.prev
    p.prev.next = p.next

    if p.prevZ:
        p.prevZ.nextZ = p.nextZ

    if p.nextZ:
        p.nextZ.prevZ = p.prevZ


def deviation(data, holeIndices, dim, triangles)->float:
    """
    Return a percentage difference between the polygon area and its triangulation area.
    Used to verify correctness of triangulation
    """
    _len = len(holeIndices)
    hasHoles = holeIndices and len(holeIndices)
    outerLen = holeIndices[0] * dim if hasHoles else len(data)

    polygonArea = abs(_polygon_area(data, 0, outerLen, dim))

    if hasHoles:
        for i in range(_len):
            start = holeIndices[i] * dim
            end = holeIndices[i + 1] * dim if i < _len - 1 else len(data)
            polygonArea -= abs(_polygon_area(data, start, end, dim))

    trianglesArea = 0.0

    for i in range(0, len(triangles), 3):
        a = triangles[i] * dim
        b = triangles[i + 1] * dim
        c = triangles[i + 2] * dim
        trianglesArea += abs(
            (data[a] - data[c]) * (data[b + 1] - data[a + 1]) -
            (data[a] - data[b]) * (data[c + 1] - data[a + 1]))

    if polygonArea == 0.0 and trianglesArea == 0.0:
        return 0

    return abs((trianglesArea - polygonArea) / polygonArea)


# 
def _triangle_area(a, b, c)->float:
    "signed area of a triangle"
    return (b[1] - a[1]) * (c[0] - b[0]) - (b[0] - a[0]) * (c[1] - b[1])

def _polygon_area(data, start, end, dim)->float:
    sum = 0.0
    j = end - dim

    for i in range(start, end, dim):
        sum += (data[j] - data[i]) * (data[i + 1] + data[j + 1])
        j = i

    return sum


# Turn a polygon in a multi-dimensional array form (e.g. as in GeoJSON) into a form Earcut accepts
def flatten(data):
    dim = len(data[0][0])
    result = {
        'vertices': [],
        'holes':    [],
        'dimensions': dim
    }
    holeIndex = 0

    for i in range(len(data)):
        for j in range(len(data[i])):
            for d in range(dim):
                result['vertices'].append(data[i][j][d])

        if i > 0:
            holeIndex += len(data[i - 1])
            result['holes'].append(holeIndex)

    return result

def unflatten(data):
    result = []

    for i in range(0, len(data), 3):
        result.append(tuple(data[i:i + 3]))

    return result



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Claudio Perez
#
import warnings

import numpy as np
Array = np.ndarray
from scipy.linalg import block_diag

from veux.model  import Model,FrameModel
from veux.state  import read_state, State, BasicState
from veux.config import Config, LineStyle, NodeStyle



def elastic_curve(x: Array, v: list, L:float)->Array:
    "compute points along Euler's elastica"
    if len(v) == 2:
        ui, uj, (vi, vj) = 0.0, 0.0, v
    else:
        ui, vi, uj, vj = v
    xi = x/L                        # local coordinate
    N1 = 1.-3.*xi**2+2.*xi**3
    N2 = L*(xi-2.*xi**2+xi**3)
    N3 = 3.*xi**2-2*xi**3
    N4 = L*(xi**3-xi**2)
    y = ui*N1 + vi*N2 + uj*N3 + vj*N4
    return y.flatten()

def elastic_tangent(x: Array, v: list, L: float)->Array:
    if len(v) == 2:
        ui, uj, (vi, vj) = 0.0, 0.0, v
    else:
        ui, vi, uj, vj = v
    xi = x/L
    M3 = 1 - xi
    M4 = 6/L*(xi-xi**2)
    M5 = 1 - 4*xi+3*xi**2
    M6 = -2*xi + 3*xi**2
    return (ui*M3 + vi*M5 + uj*M4 + vj*M6).flatten()


def displaced_profile(
        coord: Array,
        displ: Array,        #: Displacements
        vect : Array = None, #: Element orientation vector
        Q = None,
        npoints: int = 10,
        tangent: bool = False
    ):
    n = npoints
    #           (------ndm------)
    reps = 4 if len(coord[0])==3 else 2

    # 3x3 rotation into local system
    # Q = rotation(coord, vect)
    # Local displacements
    u_local = block_diag(*[Q]*reps)@displ
    # Element length
    L = np.linalg.norm(coord[-1] - coord[0])

    # longitudinal, transverse, vertical, section, elevation, plan
    li, ti, vi, si, ei, pi = u_local[:6]
    lj, tj, vj, sj, ej, pj = u_local[6:]

    Lnew  = L + lj - li
    xaxis = np.linspace(0.0, Lnew, n)

    plan_curve = elastic_curve(xaxis, [ti, pi, tj, pj], Lnew)
    elev_curve = elastic_curve(xaxis, [vi,-ei, vj,-ej], Lnew)

    local_curve = np.stack([xaxis + li, plan_curve, elev_curve])

    if tangent:
        plan_tang = elastic_tangent(xaxis, [ti, pi, tj, pj], Lnew)
        elev_tang = elastic_tangent(xaxis, [vi,-ei, vj,-ej], Lnew)

        local_tang = np.stack([np.linspace(0,0,n), plan_tang, elev_tang])
        return (
            Q.T@local_curve + coord[0][None,:].T,
            Q.T@local_tang
        )

    return Q.T@local_curve + coord[0][None,:].T


class FrameArtist:
    ndm:    int
    ndf:    int
    model:  "FrameModel"
    canvas: "Canvas"

    def __init__(self,
                 model_data,
                 ndf=None,
                 #
                 config=None,
                 # Model
                 loc=None,
                 model_config=None,
                 # Canvas
                 canvas=None,
                 #
                 **kwds):
        
        if config is None:
            config = {}

        self.config = config
        self.canvas = canvas
        vert = config.get("vertical", 2)

        self.ndm = 3

        if ndf is None:
            ndf = 6

#       elif ndf == 3:
#           self.ndm = 2

        if vert == 3:
            R = np.eye(3)
        else:
            R = np.array(((1,0, 0),
                          (0,0,-1),
                          (0,1, 0)))

        self._plot_rotation = R

        if not isinstance(model_data, Model):
            if loc is None and isinstance(model_config, dict) and "shift" in model_config:
                loc = model_config.pop("shift")

            self.model = model = FrameModel(model_data, shift=loc, rot=R,
                                            **(model_config or {}))
        else:
            self.model = model = model_data


        # Define transformation from DOFs in `state` variables to plot
        # coordinates
        if model.ndf == 1 and model.ndm == 2:
            self.dofs2plot = R@np.array(((0,),
                                         (0,),
                                         (1,)))

        elif model.ndf == 2 and model.ndm == 2:
            self.dofs2plot = R@np.array(((1, 0),
                                         (0, 1),
                                         (0, 0)))

        elif ndf == 3 and model.ndm == 2:
            self.dofs2plot = block_diag(*[R]*2)@np.array(((1,0, 0),
                                                          (0,1, 0),
                                                          (0,0, 0),

                                                          (0,0, 0),
                                                          (0,0, 0),
                                                          (0,0, 1)))

        elif ndf == 3 and model.ndm == 3:
            self.dofs2plot = block_diag(*[R]*2)@np.array(((1,0, 0),
                                                          (0,1, 0),
                                                          (0,0, 1),

                                                          (0,0, 0),
                                                          (0,0, 0),
                                                          (0,0, 0)))
        else:
            self.dofs2plot = block_diag(*[R]*2)

        self.displ_states = {}


    def _config_sketch(self,  sketch):
        strokes = {"outline", "surface", "axes", "contour", "marker", "info"}
        return {
            stroke: {
                k: v.get(stroke, None) for k, v in self.config["sketches"][sketch].items()
            } for stroke in strokes
        }

    def _add_displ_case(self, state, name=None, scale=1.0):

        if name in self.displ_states:
            self.displ_states[name].update(state)
        else:
            self.displ_states[name] = state

        return name


    def add_state(self, res_file, scale=1.0, only=None, sketch_config=None, **state_config):

        if not isinstance(res_file, (dict, Array, State)):
            state = read_state(res_file, only=only,
                               model=self.model,
                               scale=scale,
                               transform=self.dofs2plot,
                               **state_config)
        else:
            state = res_file

        # If dict of dicts, assume its a collection of responses, 
        # otherwise, just a single response
        if isinstance(state, dict) and isinstance(next(iter(state.values())), dict):
            for k, v in state.items():
                self._add_displ_case(v, name=k, scale=scale)

        elif isinstance(state, dict):
            self._add_displ_case(BasicState(state, self.model,
                                            scale=scale,
                                            transform=self.dofs2plot))

        else:
            self._add_displ_case(state, scale=scale)


    def add_point_displacements(self, displ, scale=1.0, name=None):
        displ_array = self.displ_states[name]
        for i,n in enumerate(self.model.iter_node_tags()):
            for dof in displ[n]:
                displ_array[i, dof] = 1.0

        displ_array[:,3:] *= scale/100
        displ_array[:,:3] *= scale
        return name


    # def plot_displaced_assembly(self, state=None, label=None):
    #     model = self.model
    #     N  = 10 if state is not None else 2
    #     ne = len(model["assembly"])
    #     coords = np.zeros((ne*(N+1), 3))
    #     coords.fill(np.nan)

    #     do_lines = False
    #     for i,(tag,el) in enumerate(model["assembly"].items()):
    #         if _is_frame(el) and state is not None:
    #             do_lines = True
    #             coords[(N+1)*i:(N+1)*i+N,:] = displaced_profile(model.cell_position(tag),
    #                                                             state.cell_array(tag).flatten(),
    #                                                             Q=model.frame_orientation(tag),
    #                                                             npoints=N).T
    #         elif len(el["crd"]) == 2:
    #             do_lines = True
    #             coords[(N+1)*i:(N+1)*i+N,:] = np.linspace(*el["crd"], N)

    #     if do_lines:
    #         self.canvas.plot_lines(coords[:, :self.ndm], style=LineStyle(color="red"), label=label)


    def draw_origin(self, **kwds):
        xyz = np.zeros((3,3))
        uvw = self._plot_rotation.T*kwds.get("scale", 1.0)
        off = [[0, -kwds.get("scale", 1.0)/2, 0],
               [0]*3,
               [0]*3]

        self.canvas.plot_vectors(xyz, uvw, **kwds)

#       for i,label in enumerate(kwds.get("label", [])):
#           self.canvas.annotate(label, (xyz+uvw)[i]+off[i])

    def add_elem_data(self, config=None):

        N = 3
        for type in self.model.cell_prototypes():
            name = type["name"]
            ni = len(type["instances"])
            coords = np.zeros((ni*(N+1),self.ndm))
            coords.fill(np.nan)
            for i,crd in enumerate(self.model.cell_position(type["instances"])):
                coords[(N+1)*i:(N+1)*i+N,:] = np.linspace(*crd, N)

            coords = coords.reshape(-1,4,3)[:,-3]

            self.canvas.plot_hover(coords, data=type["properties"],
                                   style=NodeStyle(shape="sphere"),
                                   label=name)

    def _draw_sketch(self, state, config):
        if config["outline"] is not None:
            self.draw_outlines(state, config=config["outline"])

        if config["surface"] is not None:
            self.draw_surfaces(state, config=config["surface"])

        if config["marker"] is not None \
            and "node" in config["marker"] \
            and config["marker"]["node"]["show"]:
            self.draw_nodes(state, config=config["marker"])

        if config["axes"] is not None:
            self.draw_axes(state, config=config["axes"])

        if "hover" in config:
            try:
                self.add_elem_data(config=config["hover"])
            except Exception as e:
                raise e
            warnings.warn(str(e))

    # sketches with or without state
    def draw_outlines(self, state=None, config=None):
        model = self.model

        if config is None:
            from veux.config import SketchConfig
            config = {type: conf["outline"] for type, conf in SketchConfig().items() if "outline" in conf}

        N = 10 if state is not None and config["frame"]["basis"] is not None else 2
        do_frames = False
        ne = sum(1 for tag in model.iter_cell_tags() if model.cell_matches(tag, "frame") or model.cell_matches(tag, "truss"))
        frames = np.zeros((ne*(N+1),3))
        frames.fill(np.nan)
#       frames = []

        quadrs = []
        trians = []
        solids = []

        for i,tag in enumerate(model.iter_cell_tags()):
            if model.cell_matches(tag, "frame") or model.cell_matches(tag, "truss"):
                if not config["frame"]["show"]:
                    continue

                if config["frame"]["basis"] is None:
                    do_frames = True
                    frames[(N+1)*i:(N+1)*i+N,:] = model.cell_position(tag, state)[[0,-1],:]
#                   frames.append( model.cell_position(tag, state)[[0,-1],:] )
                else:
                    displ = state.cell_array(tag)
                    if not hasattr(displ, "flatten"):
                        continue
                    do_frames = True
                    frames[(N+1)*i:(N+1)*i+N,:] = displaced_profile(model.cell_position(tag),
#                   frames.append (               displaced_profile(model.cell_position(tag),
                                                                    displ.flatten(),
                                                                    Q=model.frame_orientation(tag),
                                                                    npoints=N).T
#                   )
            elif model.cell_matches(tag, "plane") and config["plane"]["show"]:
                idx = model.cell_exterior(tag)
                if len(idx) == 4:
                    quadrs.append([*idx, idx[0]])
                elif len(idx) == 3:
                    trians.append([*idx, idx[0]])

            elif model.cell_matches(tag, "solid") and config["solid"]["show"]:
                # TODO: get cell faces
                idx = model.cell_exterior(tag)
                solids.append(idx)

        if do_frames and config["frame"]["show"]:
            self.canvas.plot_lines(frames[:,:self.ndm], style=config["frame"]["style"])

        if len(quadrs) > 0 and config["plane"]["show"]:
            nodes = model.node_position(state=state)
            self.canvas.plot_lines(nodes, indices=np.array(quadrs),
                                   style=config["plane"]["style"])

        if len(trians) > 0 and config["plane"]["show"]:
            nodes = model.node_position(state=state)
            self.canvas.plot_lines(nodes, indices=np.array(trians),
                                   style=config["plane"]["style"])

        if len(solids) > 0 and config["solid"]["show"]:
            nodes = model.node_position(state=state)
            self.canvas.plot_lines(nodes, indices=np.array(solids),
                                          style=config["solid"]["style"])

    def draw_surfaces(self, state=None, field=None, layer=None, config=None):
        model = self.model

        if config is None:
            from veux.config import SketchConfig
            config = {type: conf["surface"] for type, conf in SketchConfig().items() if "surface" in conf}
        # Draw extruded frames
        from veux.frame import extrude
        if "frame" in config and config["frame"]["show"]:
            extrude.draw_extrusions(model,
                                    canvas=self.canvas,
                                    state=state,
                                    config=config["frame"])

        # Draw filled mesh for cell-like elements
        triangles = []
        if "plane" in config and config["plane"]["show"]:
            nodes = model.node_position(state=state)

            for tag in model.iter_cell_tags():
                if not model.cell_matches(tag, "frame"):
                    triangles.extend(model.cell_triangles(tag))

        if len(triangles) > 0:
            mesh = self.canvas.plot_mesh(nodes, np.array(triangles), style=config["plane"]["style"])

            if field is not None:
                if isinstance(field, dict):
                    field = np.array([field[node] for node in model.iter_node_tags()])
                self.canvas.plot_mesh_field(mesh, field=field)

        return

    def draw_nodes(self, state=None, data=None, label=None, config=None):
        if state is not None and state.rotation is not None and self.model.ndm == 3:
            rotations = self.model.node_rotation(state=state)
        else:
            rotations = None

        coord = self.model.node_position(state=state)
        self.canvas.plot_nodes(coord[:,:self.ndm], label=label,
                               rotations=rotations,
                               style=config["node"]["style"])

        if state is None:
            self.canvas.plot_hover(coord[:,:self.ndm],
                                   label="node",
                                   keys=["tag", "crd"],
                                   data=[[str(k), list(map(str, coord[i]))]
                                       for i,k in enumerate(self.model.iter_node_tags())])

    def draw_axes(self, state=None, config=None):
        ne = sum(1 for tag in self.model.iter_cell_tags() if self.model.cell_matches(tag, "frame"))
        xyz, uvw = np.nan*np.zeros((2, ne, 3, 3))

        for i,tag in enumerate(self.model.iter_cell_tags()):

            axes = self.model.frame_orientation(tag)
            if axes is None or not config["frame"]["show"]:
                continue
            crd = self.model.cell_position(tag, state=state) #el["crd"]
            scale = np.linalg.norm(crd[-1] - crd[0])/10
            coord = sum(i for i in crd)/len(self.model.cell_indices(tag))
            xyz[i,:,:] = np.array([coord]*3)
            uvw[i,:,:] = scale*axes

        self.canvas.plot_vectors(xyz.reshape(ne*3,3), uvw.reshape(ne*3,3))


# PUBLIC

    def draw(self):
        # Background
        default = self.config["sketches"]["default"]
        if "origin" in default and default["origin"]["axes"]["show"]:
            origin = default["origin"]
            self.draw_origin(line_style=origin["axes"]["style"],
                                  scale=origin["axes"]["scale"],
                                  label=origin["axes"]["label"])

        # Reference
        if ("reference" in self.config["sketches"] \
                and self.config["sketches"]["reference"]):
            self._draw_sketch(None, config=self._config_sketch("reference"))

        elif len(self.displ_states) == 0:
            self._draw_sketch(None, config=self._config_sketch("default"))

        # Deformations
        for layer, state in self.displ_states.items():
            self._draw_sketch(config=self._config_sketch("displaced"),
                        state=state)

        self.canvas.build()
        return self

    def save(self, filename):
        self.canvas.write(filename)

    def repl(self):
        from opensees.repl.__main__ import OpenSeesREPL
        self.canvas.plt.ion()

        try:
            from IPython import get_ipython
            get_ipython().run_magic_line('matplotlib')
        except:
            pass

        repl = OpenSeesREPL()

        def plot(*args):
            if len(args) == 0:
                return self.draw()

            elif hasattr(self, "plot_"+args[0]):
                return getattr(self, "plot_"+args[0])(*args[1:])

            elif hasattr(self, args[0]):
                return getattr(self, args[0])(*args[1:])

        repl.interp._interp.createcommand("plot", plot)
        # repl.interp._interp.createcommand("show", lambda *args: self.canvas.show())
        repl.repl()




# Claudio Perez
import sys
import warnings
import numpy as np
from scipy.spatial.transform import Rotation

import shps.curve

import veux
import veux.frame
from veux.utility.earcut import earcut
from veux.model import read_model
from veux.errors import RenderError
from veux.config import MeshStyle

import sys
import warnings
import numpy as np
from scipy.spatial.transform import Rotation

import shps.curve

import veux
import veux.frame
from veux.utility.earcut import earcut
from veux.model import read_model
from veux.errors import RenderError
from veux.config import MeshStyle

# 1) Import the new Extrusion class
from shps.frame.extrude import Extrusion

def draw_extrusions2(model, canvas, state=None, config=None):
    """
    Draw extruded frame elements into 'canvas' as a simple mesh (side faces + end caps),
    optionally applying displacements/rotations from 'state'.

    'Extrusion' handles building the side + cap faces in local coords, then we map them 
    to global coords using each ring’s X[j] + R[j].

    The final mesh is plotted with 'canvas.plot_mesh'. Then, if config["outline"] 
    includes 'tran' or 'long', we also call 'canvas.plot_lines' for edges or cross-hatching.
    """
    ndm = 3

    if config is None:
        config = {"style": MeshStyle(color="gray")}
    scale_section = config.get("scale", 1.0)

    #----------------------------------------------------------
    # 1) Build local geometry (side + caps) using Extrusion
    #----------------------------------------------------------
    # This extracts the ring-building logic from your original code
    extr = Extrusion(model, scale=scale_section, do_end_caps=True)
    local_positions = extr.vertices()    # Nx3 (local coords)
    triangles       = extr.triangles()   # Mx3
    ring_ranges     = extr.ring_ranges() # [((elem_name,j), start_idx, end_idx), ...]

    # If there's no geometry, bail out
    if len(triangles) == 0:
        return

    # Create a global array for final coordinates
    coords = np.zeros_like(local_positions, dtype=float)

    # We'll also build 'locoor' for local texture or param 
    # for each vertex. The original code sets them in a complicated way, 
    # but here we’ll replicate the same shape & fill them with something 
    # similar to your old approach.
    locoor = [None]*len(local_positions)

    #----------------------------------------------------------
    # 2) For each ring, apply the global transform X[j] + R[j]@local_pt
    #    from either the reference config or the 'state'
    #----------------------------------------------------------
    for ((elem_name, j), start_idx, end_idx) in ring_ranges:
        el  = model["assembly"][elem_name]
        nen = len(el["nodes"])

        # Original coords for this element
        X_ref = np.array(el["crd"])  # shape (nen, 3)

        if state is not None:
            # Displacements
            glob_displ = state.cell_array(elem_name, state.position)
            # returns a (3, nen) array.
            X_def = shps.curve.displace(X_ref, glob_displ, nen).T

            R_all = state.cell_array(elem_name, state.rotation)  # list of 3x3 or (nen, 3x3)
        else:
            # If no state, just do reference
            X_def = X_ref.T  # shape (3, nen)
            R_all = [model.frame_orientation(elem_name).T]*nen

        # For ring j, we have a 3x1 translation = X_def[:, j]
        # and a rotation R_all[j], which is 3x3
        trans_j = X_def[:, j]
        rot_j   = R_all[j]

        # For each vertex i in [start_idx, end_idx), transform
        for i in range(start_idx, end_idx):
            local_pt = local_positions[i]      # shape (3,)
            # Global coords
            coords[i] = trans_j + rot_j @ local_pt

            # locoor? The original code sets something like
            #   locoor.append([ (j+0)/nen+0.1, 0.1+(k+0)/(noe+0 ) ])
            # But we no longer track the "k" directly, so let's store a placeholder.
            # Or we can store extr.texcoords() if you do that in your Extrusion class.
            locoor[i] = [(j+0)/max(nen,1)+0.1, 0.1]  # minimal

    #----------------------------------------------------------
    # 3) Reverse triangle winding
    #----------------------------------------------------------
    triang = [list(reversed(face)) for face in triangles]

    #----------------------------------------------------------
    # 4) Plot the main mesh
    #----------------------------------------------------------
    canvas.plot_mesh(coords, triang, local_coords=locoor, style=config["style"])

    if "outline" not in config:
        return

    #----------------------------------------------------------
    # 5) Draw edges
    #----------------------------------------------------------
    nan = np.array([0,0,0], dtype=float)*np.nan
    IDX = np.array(((0,2),(0,1)))  # from your code

    if "tran" in config["outline"]:
        # Then we create tri_points in the same way your code does:
        #   tri_points = [ coords[idx] if (j+1)%3 else nan for j,idx in enumerate(np.array(triang).reshape(-1)) ]
        tri_flat = np.array(triang).reshape(-1)
        tri_points = []
        for j, idx in enumerate(tri_flat):
            # after each 3rd vertex, we insert a nan
            tri_points.append(coords[idx])
            if (j+1)%3 == 0:
                tri_points.append(nan)
        tri_points = np.array(tri_points)
        canvas.plot_lines(tri_points, style=config.get("line_style", None))

    elif "long" in config["outline"]:
        # replicate that logic for "long"
        # tri_points = [ coords[i] if j%2 else nan for j,face in enumerate(triang) for i in face[IDX[j%2]] ]
        # but your code also checks "if j not in no_outline"? We'll omit that since it's the original logic 
        # for skipping edges on big outlines. If you truly want that, we can do it the same way.
        tri_points = []
        tri_array  = np.array(triang)
        for j, face in enumerate(tri_array):
            # each face is [i0,i1,i2]
            # pick out face[IDX[j%2]] => [face[idx0], face[idx1]]
            # then for each i in that pair, we do coords[i], unless we do the 'nan' row
            if j%2 == 0:
                # put a nan row to separate from previous?
                tri_points.append(nan)
            i0 = face[IDX[j%2][0]]
            i1 = face[IDX[j%2][1]]
            tri_points.append(coords[i0])
            tri_points.append(coords[i1])

        tri_points = np.array(tri_points)
        canvas.plot_lines(tri_points, style=config.get("line_style", None))

    # done!
    return

def draw_extrusions(model, canvas, state=None, config=None):
    #
    #     x-------o---------o---------o
    #   /       /         /
    # x--------o<--------o---------o
    # |        |       / ^
    # |        |     /   |
    # |        |   /     |
    # |        | /       |
    # x--------o-------->o---------o
    #
    ndm = 3

    coords = [] # Global mesh coordinates
    triang = [] # Triangle indices into coords
    caps   = []
    locoor = [] # Local mesh coordinates, used for textures

    if config is None:
        config = {
                "style": MeshStyle(color="gray")
        }
    scale_section = config["scale"]


    I = 0
    # Track outlines with excessive edges (eg, circles) to later avoid showing
    # their edges
    no_outline = set()
    for i,el in enumerate(model["assembly"].values()):
        tag  = el["name"]

        outline = model.cell_section(el["name"])
        if outline is None:
            continue

        outline_scale = scale_section

        nen  = len(el["nodes"])

        noe = len(outline) # number of outline edges
        if state is not None:
            glob_displ = state.cell_array(tag, state.position)
            X = shps.curve.displace(el["crd"], glob_displ, nen).T
            R = state.cell_array(tag, state.rotation)
        else:
            outline = outline*0.99
            outline_scale *= 0.99
            X = np.array(el["crd"])
            R = [model.frame_orientation(tag).T]*nen


        try:
            caps.append(I+np.array(earcut(model.cell_section(tag, 0)[:,1:])))
            caps.append(I+(nen-1)*noe + np.array(earcut(model.cell_section(tag, 1)[:,1:])))
        except Exception as e:
            warnings.warn(f"Earcut failed with message: {e}")


        # Loop over sample points along element length to assemble
        # `coord` and `triang` arrays
        for j in range(nen):
            outline = model.cell_section(tag, j).copy() # TODO: Pass float between 0 and 1 instead of j
            outline[:,1:] *= outline_scale
            # Loop over section edges
            for k,edge in enumerate(outline):
                # Append rotated section coordinates to list of coordinates
                coords.append(X[j, :] + R[j]@edge)
                locoor.append(
                             [ (j+0)/nen+0.1,  0.1+(k+0)/(noe+0) ]
                )


                if j == 0:
                    # Skip the first section
                    continue

                elif k < noe-1:
                    triang.extend([
                        # Tie two triangles to this edge
                        [I+    noe*j + k,   I+    noe*j + k + 1,    I+noe*(j-1) + k],
                        [I+noe*j + k + 1,   I+noe*(j-1) + k + 1,    I+noe*(j-1) + k]
                    ])
                else:
                    # elif j < N-1:
                    triang.extend([
                        [I+    noe*j + k,    I + noe*j , I+noe*(j-1) + k],
                        [      I + noe*j, I + noe*(j-1), I+noe*(j-1) + k]
                    ])

                if len(outline) > 25:
                    no_outline.add(len(triang)-1)
                    no_outline.add(len(triang)-2)

        I += nen*noe

    triang = [list(reversed(i)) for i in triang]

    if len(triang) == 0:
        return

    mesh = canvas.plot_mesh(coords, triang, local_coords=locoor, style=config["style"])

    if len(caps) > 0:
        for cap in caps:
            canvas.plot_mesh(mesh.vertices, cap, style=config["style"])


    IDX = np.array((
        (0, 2),
        (0, 1)
    ))

    triang = [list(reversed(i)) for i in triang]

    nan = np.zeros(ndm)*np.nan
    coords = np.array(coords)
    if "tran" in config["outline"]:
        tri_points = np.array([
            coords[idx]  if (j+1)%3 else nan
            for j,idx in enumerate(np.array(triang).reshape(-1))
        ])
    elif "long" in config["outline"]:
        tri_points = np.array([
            coords[i]  if j%2 else nan
            for j,idx in enumerate(np.array(triang)) for i in idx[IDX[j%2]] if j not in no_outline
        ])
    else:
        return

    canvas.plot_lines(tri_points,
                      style=config["line_style"]
    )

class so3:
    @classmethod
    def exp(cls, vect):
        return Rotation.from_rotvec(vect).as_matrix()

def _add_moment(artist, loc, axis):
    import meshio
    mesh_data = meshio.read(veux.assets/'chrystals_moment.stl')
    coords = mesh_data.points

    coords = np.einsum('ik, kj -> ij',  coords,
                       so3.exp([0, 0, -np.pi/4])@so3.exp(axis))
    coords = 1e-3*coords + loc
#   for node in coords:
#       node = so3.exp(axis)@node
    for i in mesh_data.cells:
        if i.type == "triangle":
            triangles =  i.data #mesh_data.cells['triangle']
            break

    artist.canvas.plot_mesh(coords, triangles)


def _render(sam_file, res_file=None, **opts):
    # Configuration is determined by successively layering
    # from sources with the following priorities:
    #      defaults < file configs < kwds 

    config = veux.config.Config()


    if sam_file is None:
        raise RenderError("Expected positional argument <sam-file>")

    # Read and clean model
    if not isinstance(sam_file, dict):
        model = read_model(sam_file)
    else:
        model = sam_file

    if "RendererConfiguration" in model:
        veux.apply_config(model["RendererConfiguration"], config)

    veux.apply_config(opts, config)

    artist = veux.FrameArtist(model, **config)

    draw_extrusions(artist.model, artist.canvas, config=opts)

    # -----------------------------------------------------------

    soln = veux.state.read_state(res_file, artist.model, **opts)
    if soln is not None:
        if "time" not in opts:
            soln = soln[soln.times[-1]]

        draw_extrusions(artist.model, artist.canvas, soln, opts)
        # -----------------------------------------------------------
        _add_moment(artist,
                    loc  = [1.0, 0.0, 0.0],
                    axis = [0, np.pi/2, 0])
        # -----------------------------------------------------------

    artist.draw()
    return artist


if __name__ == "__main__":
    import veux.parser
    config = veux.parser.parse_args(sys.argv)

    try:
        artist = _render(**config)

        # write plot to file if output file name provided
        if config["write_file"]:
            artist.save(config["write_file"])


        # Otherwise either create popup, or start server
        elif hasattr(artist.canvas, "popup"):
            artist.canvas.popup()

        elif hasattr(artist.canvas, "to_glb"):
            import veux.server
            server = veux.server.Server(glb=artist.canvas.to_glb(),
                                        viewer=config["viewer_config"].get("name", None))
            server.run(config["server_config"].get("port", None))

        elif hasattr(artist.canvas, "to_html"):
            import veux.server
            server = veux.server.Server(html=artist.canvas.to_html())
            server.run(config["server_config"].get("port", None))

    except (FileNotFoundError, RenderError) as e:
        # Catch expected errors to avoid printing an ugly/unnecessary stack trace.
        print(e, file=sys.stderr)
        print("         Run '{NAME} --help' for more information".format(NAME=sys.argv[0]), file=sys.stderr)
        sys.exit()




import numpy as np 


def quad8n_val_at_center(vals):
    """
    Calculate values at the center of 8-node quad element.
    """

    val_c1 = -np.mean(vals[[0, 1, 2, 3]], axis=0) + 2*np.mean(vals[[4, 5, 6, 7]], axis=0)  # noqa: E501
    # val_c1 = -np.sum(vals[[0, 1, 2, 3]], axis=0)/4 + np.sum(vals[[4, 5, 6, 7]], axis=0)/2  # noqa: E501
    # val_c2 = -np.sum(vals[[0, 1, 2, 3]], axis=0)/4 + np.sum(vals[[4, 5, 6, 7]], axis=0)/2  # noqa: E501
    # val_c3 = np.mean(vals[[4, 5, 6, 7]], axis=0)
    # val_c4 = np.mean(vals[[0, 1, 2, 3]], axis=0)
    # val_c5 = np.mean(vals, axis=0)
    # return val_c1, val_c2, val_c3, val_c4, val_c5

    return val_c1


def quads_to_4tris(quads_conn, nds_crd, nds_val):
    """
    Get triangles connectivity

    Notes:
        Triangles connectivity array is based on
        quadrilaterals connectivity.
        Each quad is split into four triangles.
        New nodes are created at the quad centroid.

    See also:
        function: quads_to_8tris_9n, quads_to_8tris_8n
    """
    n_quads, _ = quads_conn.shape
    n_nds, _ = nds_crd.shape

    tris_conn = np.zeros((4*n_quads, 3), dtype=int)

    for i, quad_conn in enumerate(quads_conn):
        j = 4*i
        n0, n1, n2, n3 = quad_conn

        # triangles connectivity
        tris_conn[j  ] = np.array([n0, n1, n_nds+i])
        tris_conn[j+1] = np.array([n1, n2, n_nds+i])
        tris_conn[j+2] = np.array([n2, n3, n_nds+i])
        tris_conn[j+3] = np.array([n3, n0, n_nds+i])

    return tris_conn, None, None


# see also quads_to_8tris_9n
def quads_to_8tris_8n(quads_conn, nds_crd, nds_val):
    """
    Get triangles connectivity, coordinates and new values at quad centroids.

    Notes:
        Triangles connectivity array is based on
        quadrilaterals connectivity.
        Each quad is split into eight triangles.
        New nodes are created at the quad centroid.

    See also:
        function: quads_to_8tris_9n, quads_to_4tris
    """
    n_quads, _ = quads_conn.shape
    n_nds, _ = nds_crd.shape

    # coordinates and values at quad centroids _c_
    nds_c_crd = np.zeros((n_quads, 2))
    nds_c_val = np.zeros(n_quads)

    tris_conn = np.zeros((8*n_quads, 3), dtype=int)

    for i, quad_conn in enumerate(quads_conn):
        j = 8*i
        n0, n1, n2, n3, n4, n5, n6, n7 = quad_conn

        # quad centroids
        # nds_c_crd[i] = np.array([np.sum(nds_crd[[n0, n1, n2, n3], 0])/4.,
        #                          np.sum(nds_crd[[n0, n1, n2, n3], 1])/4.])
        # nds_c_val[i] = np.sum(nds_val[[n0, n1, n2, n3]])/4.
        nds_c_crd[i] = quad8n_val_at_center(nds_crd[[n0, n1, n2, n3,
                                                     n4, n5, n6, n7]])
        nds_c_val[i] = quad8n_val_at_center(nds_val[[n0, n1, n2, n3,
                                                     n4, n5, n6, n7]])

        # triangles connectivity
        tris_conn[j  ] = np.array([n0, n4, n_nds+i])
        tris_conn[j+1] = np.array([n4, n1, n_nds+i])
        tris_conn[j+2] = np.array([n1, n5, n_nds+i])
        tris_conn[j+3] = np.array([n5, n2, n_nds+i])
        tris_conn[j+4] = np.array([n2, n6, n_nds+i])
        tris_conn[j+5] = np.array([n6, n3, n_nds+i])
        tris_conn[j+6] = np.array([n3, n7, n_nds+i])
        tris_conn[j+7] = np.array([n7, n0, n_nds+i])

    return tris_conn, nds_c_crd, nds_c_val


# see also quads_to_8tris_8n
def quads_to_8tris_9n(quads_conn):
    """
    Get triangles connectivity, coordinates and new values at quad centroids.

    Args:
        quads_conn (ndarray):

    Returns:
        tris_conn, nds_c_crd, nds_c_val (tuple):

    Notes:
        Triangles connectivity array is based on
        quadrilaterals connectivity.
        Each quad is split into eight triangles.
        New nodes are created at the quad centroid.
    """
    n_quads, _ = quads_conn.shape

    tris_conn = np.zeros((8*n_quads, 3), dtype=int)

    for i, quad_conn in enumerate(quads_conn):
        j = 8*i
        n0, n1, n2, n3, n4, n5, n6, n7, n8 = quad_conn

        # triangles connectivity
        tris_conn[j]   = np.array([n0, n4, n8])
        tris_conn[j+1] = np.array([n4, n1, n8])
        tris_conn[j+2] = np.array([n1, n5, n8])
        tris_conn[j+3] = np.array([n5, n2, n8])
        tris_conn[j+4] = np.array([n2, n6, n8])
        tris_conn[j+5] = np.array([n6, n3, n8])
        tris_conn[j+6] = np.array([n3, n7, n8])
        tris_conn[j+7] = np.array([n7, n0, n8])

    return tris_conn


def bricks_to_24tris(bricks_conn, nds_crd, nds_val, disps=None):
    """
    Get triangles connectivity, coordinates and new vals at brick face centroids.

    Notes:
        Triangles connectivity array is based on
        stdBricks connectivity.
        Each of 6 brick faces is split into four triangles.
        New nodes are created at the face centroid.

    See also:
        function: bricks_to_8tris_9n, bricks_to_8tris_8n
    """
    n_bricks, _ = bricks_conn.shape
    n_nds, _ = nds_crd.shape

    # coordinates and values at brick centroids _c_
    nds_c_crd = np.zeros((n_bricks*6, 3))
    nds_c_val = np.zeros(n_bricks*6)

    disps_c = None
    if disps is not None:
        disps_c = np.zeros((n_bricks*6, 3))

    tris_conn = np.zeros((24*n_bricks, 3), dtype=int)

    for i, brick_conn in enumerate(bricks_conn):
        j = 24*i
        n0, n1, n2, n3, n4, n5, n6, n7 = brick_conn

        # brick centroids
        nds_c_crd[i*6] = np.array([np.sum(nds_crd[[n0, n1, n5, n4], 0])/4.,
                                   np.sum(nds_crd[[n0, n1, n5, n4], 1])/4.,
                                   np.sum(nds_crd[[n0, n1, n5, n5], 2])/4.])
        nds_c_crd[i*6+1] = np.array([np.sum(nds_crd[[n1, n2, n6, n5], 0])/4.,
                                     np.sum(nds_crd[[n1, n2, n6, n5], 1])/4.,
                                     np.sum(nds_crd[[n1, n2, n6, n5], 2])/4.])
        nds_c_crd[i*6+2] = np.array([np.sum(nds_crd[[n2, n3, n7, n6], 0])/4.,
                                     np.sum(nds_crd[[n2, n3, n7, n6], 1])/4.,
                                     np.sum(nds_crd[[n2, n3, n7, n6], 2])/4.])
        nds_c_crd[i*6+3] = np.array([np.sum(nds_crd[[n3, n0, n4, n7], 0])/4.,
                                     np.sum(nds_crd[[n3, n0, n4, n7], 1])/4.,
                                     np.sum(nds_crd[[n3, n0, n4, n7], 2])/4.])
        nds_c_crd[i*6+4] = np.array([np.sum(nds_crd[[n4, n5, n6, n7], 0])/4.,
                                     np.sum(nds_crd[[n4, n5, n6, n7], 1])/4.,
                                     np.sum(nds_crd[[n4, n5, n6, n7], 2])/4.])
        nds_c_crd[i*6+5] = np.array([np.sum(nds_crd[[n0, n1, n2, n3], 0])/4.,
                                     np.sum(nds_crd[[n0, n1, n2, n3], 1])/4.,
                                     np.sum(nds_crd[[n0, n1, n2, n3], 2])/4.])

        nds_c_val[6*i  ] = np.sum(nds_val[[n0, n1, n5, n4]])/4.
        nds_c_val[6*i+1] = np.sum(nds_val[[n1, n2, n6, n5]])/4.
        nds_c_val[6*i+2] = np.sum(nds_val[[n2, n3, n7, n6]])/4.
        nds_c_val[6*i+3] = np.sum(nds_val[[n3, n0, n4, n7]])/4.
        nds_c_val[6*i+4] = np.sum(nds_val[[n4, n5, n6, n7]])/4.
        nds_c_val[6*i+5] = np.sum(nds_val[[n0, n1, n2, n3]])/4.

        # triangles connectivity
        tris_conn[j  ] = np.array([n0, n1, n_nds+i*6])
        tris_conn[j+1] = np.array([n1, n5, n_nds+i*6])
        tris_conn[j+2] = np.array([n5, n4, n_nds+i*6])
        tris_conn[j+3] = np.array([n4, n0, n_nds+i*6])

        tris_conn[j+4] = np.array([n1, n2, n_nds+i*6+1])
        tris_conn[j+5] = np.array([n2, n6, n_nds+i*6+1])
        tris_conn[j+6] = np.array([n6, n5, n_nds+i*6+1])
        tris_conn[j+7] = np.array([n5, n1, n_nds+i*6+1])

        tris_conn[j+8] = np.array([n2, n3, n_nds+i*6+2])
        tris_conn[j+9] = np.array([n3, n7, n_nds+i*6+2])
        tris_conn[j+10] = np.array([n7, n6, n_nds+i*6+2])
        tris_conn[j+11] = np.array([n6, n2, n_nds+i*6+2])

        tris_conn[j+12] = np.array([n3, n0, n_nds+i*6+3])
        tris_conn[j+13] = np.array([n0, n4, n_nds+i*6+3])
        tris_conn[j+14] = np.array([n4, n7, n_nds+i*6+3])
        tris_conn[j+15] = np.array([n7, n3, n_nds+i*6+3])

        tris_conn[j+16] = np.array([n4, n5, n_nds+i*6+4])
        tris_conn[j+17] = np.array([n5, n6, n_nds+i*6+4])
        tris_conn[j+18] = np.array([n6, n7, n_nds+i*6+4])
        tris_conn[j+19] = np.array([n7, n4, n_nds+i*6+4])

        tris_conn[j+20] = np.array([n0, n1, n_nds+i*6+5])
        tris_conn[j+21] = np.array([n1, n2, n_nds+i*6+5])
        tris_conn[j+22] = np.array([n2, n3, n_nds+i*6+5])
        tris_conn[j+23] = np.array([n3, n0, n_nds+i*6+5])

        if disps is not None:
            disps_c[6*i] = np.sum(disps[[n0, n1, n5, n4]], axis=0)/4.
            disps_c[6*i+1] = np.sum(disps[[n1, n2, n6, n5]], axis=0)/4.
            disps_c[6*i+2] = np.sum(disps[[n2, n3, n7, n6]], axis=0)/4.
            disps_c[6*i+3] = np.sum(disps[[n3, n0, n4, n7]], axis=0)/4.
            disps_c[6*i+4] = np.sum(disps[[n4, n5, n6, n7]], axis=0)/4.
            disps_c[6*i+5] = np.sum(disps[[n0, n1, n2, n3]], axis=0)/4.

    return tris_conn, nds_c_crd, nds_c_val, disps_c


# brick20n bricks to tris
def bricks_to_48tris(bricks_conn, nds_crd, nds_val, disps=None):
    """
    Get triangles connectivity, coordinates and new vals at brick face centroids,
    for brick20n

    Notes:
        Triangles connectivity array is based on
        stdBricks connectivity.
        Each of 6 brick faces is split into four triangles.
        New nodes are created at the face centroid.

    See also:
        function: bricks_to_24tris
    """
    n_bricks, _ = bricks_conn.shape
    n_nds, _ = nds_crd.shape

    # coordinates and values at brick centroids _c_
    nds_c_crd = np.zeros((n_bricks*6, 3))
    nds_c_val = np.zeros(n_bricks*6)

    disps_c = None
    if disps is not None:
        disps_c = np.zeros((n_bricks*6, 3))

    tris_conn = np.zeros((48*n_bricks, 3), dtype=int)

    for i, brick_conn in enumerate(bricks_conn):
        j = 48*i
        n0, n1, n2, n3, n4, n5, n6, n7, n8, n9, n10, n11, n12, n13, \
            n14, n15, n16, n17, n18, n19 = brick_conn

        # brick centroids
        nds_c_crd[i*6] = np.array([np.sum(nds_crd[[n0, n1, n5, n4], 0])/4.,
                                   np.sum(nds_crd[[n0, n1, n5, n4], 1])/4.,
                                   np.sum(nds_crd[[n0, n1, n5, n5], 2])/4.])
        nds_c_crd[i*6+1] = np.array([np.sum(nds_crd[[n1, n2, n6, n5], 0])/4.,
                                     np.sum(nds_crd[[n1, n2, n6, n5], 1])/4.,
                                     np.sum(nds_crd[[n1, n2, n6, n5], 2])/4.])
        nds_c_crd[i*6+2] = np.array([np.sum(nds_crd[[n2, n3, n7, n6], 0])/4.,
                                     np.sum(nds_crd[[n2, n3, n7, n6], 1])/4.,
                                     np.sum(nds_crd[[n2, n3, n7, n6], 2])/4.])
        nds_c_crd[i*6+3] = np.array([np.sum(nds_crd[[n3, n0, n4, n7], 0])/4.,
                                     np.sum(nds_crd[[n3, n0, n4, n7], 1])/4.,
                                     np.sum(nds_crd[[n3, n0, n4, n7], 2])/4.])
        nds_c_crd[i*6+4] = np.array([np.sum(nds_crd[[n4, n5, n6, n7], 0])/4.,
                                     np.sum(nds_crd[[n4, n5, n6, n7], 1])/4.,
                                     np.sum(nds_crd[[n4, n5, n6, n7], 2])/4.])
        nds_c_crd[i*6+5] = np.array([np.sum(nds_crd[[n0, n1, n2, n3], 0])/4.,
                                     np.sum(nds_crd[[n0, n1, n2, n3], 1])/4.,
                                     np.sum(nds_crd[[n0, n1, n2, n3], 2])/4.])

        nds_c_val[6*i  ] = np.sum(nds_val[[n0, n1, n5, n4]])/4.
        nds_c_val[6*i+1] = np.sum(nds_val[[n1, n2, n6, n5]])/4.
        nds_c_val[6*i+2] = np.sum(nds_val[[n2, n3, n7, n6]])/4.
        nds_c_val[6*i+3] = np.sum(nds_val[[n3, n0, n4, n7]])/4.
        nds_c_val[6*i+4] = np.sum(nds_val[[n4, n5, n6, n7]])/4.
        nds_c_val[6*i+5] = np.sum(nds_val[[n0, n1, n2, n3]])/4.

        # triangles connectivity
        tris_conn[j  ] = np.array([n0, n8, n_nds+i*6])
        tris_conn[j+1] = np.array([n8, n1, n_nds+i*6])
        tris_conn[j+2] = np.array([n1, n17, n_nds+i*6])
        tris_conn[j+3] = np.array([n17, n5, n_nds+i*6])
        tris_conn[j+4] = np.array([n5, n12, n_nds+i*6])
        tris_conn[j+5] = np.array([n12, n4, n_nds+i*6])
        tris_conn[j+6] = np.array([n4, n16, n_nds+i*6])
        tris_conn[j+7] = np.array([n16, n0, n_nds+i*6])

        tris_conn[j+8] = np.array([n1, n9, n_nds+i*6+1])
        tris_conn[j+9] = np.array([n9, n2, n_nds+i*6+1])
        tris_conn[j+10] = np.array([n2, n18, n_nds+i*6+1])
        tris_conn[j+11] = np.array([n18, n6, n_nds+i*6+1])
        tris_conn[j+12] = np.array([n6, n13, n_nds+i*6+1])
        tris_conn[j+13] = np.array([n13, n5, n_nds+i*6+1])
        tris_conn[j+14] = np.array([n5, n17, n_nds+i*6+1])
        tris_conn[j+15] = np.array([n17, n1, n_nds+i*6+1])

        tris_conn[j+16] = np.array([n2, n10, n_nds+i*6+2])
        tris_conn[j+17] = np.array([n10, n3, n_nds+i*6+2])
        tris_conn[j+18] = np.array([n3, n19, n_nds+i*6+2])
        tris_conn[j+19] = np.array([n19, n7, n_nds+i*6+2])
        tris_conn[j+20] = np.array([n7, n14, n_nds+i*6+2])
        tris_conn[j+21] = np.array([n14, n6, n_nds+i*6+2])
        tris_conn[j+22] = np.array([n6, n18, n_nds+i*6+2])
        tris_conn[j+23] = np.array([n18, n2, n_nds+i*6+2])

        tris_conn[j+24] = np.array([n3, n11, n_nds+i*6+3])
        tris_conn[j+25] = np.array([n11, n0, n_nds+i*6+3])
        tris_conn[j+26] = np.array([n0, n16, n_nds+i*6+3])
        tris_conn[j+27] = np.array([n16, n4, n_nds+i*6+3])
        tris_conn[j+28] = np.array([n4, n15, n_nds+i*6+3])
        tris_conn[j+29] = np.array([n15, n7, n_nds+i*6+3])
        tris_conn[j+30] = np.array([n7, n19, n_nds+i*6+3])
        tris_conn[j+31] = np.array([n19, n3, n_nds+i*6+3])

        tris_conn[j+32] = np.array([n4, n12, n_nds+i*6+4])
        tris_conn[j+33] = np.array([n12, n5, n_nds+i*6+4])
        tris_conn[j+34] = np.array([n5, n13, n_nds+i*6+4])
        tris_conn[j+35] = np.array([n13, n6, n_nds+i*6+4])
        tris_conn[j+36] = np.array([n6, n14, n_nds+i*6+4])
        tris_conn[j+37] = np.array([n14, n7, n_nds+i*6+4])
        tris_conn[j+38] = np.array([n7, n15, n_nds+i*6+4])
        tris_conn[j+39] = np.array([n15, n4, n_nds+i*6+4])

        tris_conn[j+40] = np.array([n0, n8, n_nds+i*6+5])
        tris_conn[j+41] = np.array([n8, n1, n_nds+i*6+5])
        tris_conn[j+42] = np.array([n1, n9, n_nds+i*6+5])
        tris_conn[j+43] = np.array([n9, n2, n_nds+i*6+5])
        tris_conn[j+44] = np.array([n2, n10, n_nds+i*6+5])
        tris_conn[j+45] = np.array([n10, n3, n_nds+i*6+5])
        tris_conn[j+46] = np.array([n3, n11, n_nds+i*6+5])
        tris_conn[j+47] = np.array([n11, n0, n_nds+i*6+5])

        if disps is not None:
            disps_c[6*i] = np.sum(disps[[n0, n1, n5, n4,
                                         n8, n17, n12, n16]], axis=0)/4.
            disps_c[6*i+1] = np.sum(disps[[n1, n2, n6, n5,
                                           n9, n18, n13, n17]], axis=0)/4.
            disps_c[6*i+2] = np.sum(disps[[n2, n3, n7, n6,
                                           n10, n19, n14, n18]], axis=0)/4.
            disps_c[6*i+3] = np.sum(disps[[n3, n0, n4, n7,
                                           n11, n16, n15, n19]], axis=0)/4.
            disps_c[6*i+4] = np.sum(disps[[n4, n5, n6, n7,
                                           n12, n13, n14, n15]], axis=0)/4.
            disps_c[6*i+5] = np.sum(disps[[n0, n1, n2, n3,
                                           n8, n9, n10, n11]], axis=0)/4.

    return tris_conn, nds_c_crd, nds_c_val, disps_c


def tetra4n_to_4tris(tetras4n_conn):
    """Get triangles connectivity.

    Four-node tetrahedron is subdivided into four triangles

    Args:
        tetra4n_conn (ndarray):

    Returns:
        tris_conn_subdiv (ndarray):
    """
    n_tetras, _ = tetras4n_conn.shape
    # n_nds, _ = nds_crd.shape

    tris_conn_subdiv = np.zeros((4*n_tetras, 3), dtype=int)

    for i, tetra4n_conn in enumerate(tetras4n_conn):
        j = 4*i
        n0, n1, n2, n3 = tetra4n_conn

        # triangles connectivity
        tris_conn_subdiv[j] = np.array([n0, n1, n2])
        tris_conn_subdiv[j+1] = np.array([n0, n3, n1])
        tris_conn_subdiv[j+2] = np.array([n0, n2, n3])
        tris_conn_subdiv[j+3] = np.array([n1, n2, n3])

    return tris_conn_subdiv


def tetra10n_to_16tris(tetras10n_conn):
    """Get triangles connectivity.

    Six-node triangle is subdivided into four triangles

    Args:
        tetra10n_conn (ndarray):

    Returns:
        tris_conn_subdiv (ndarray):
    """
    n_tetras, _ = tetras10n_conn.shape
    # n_nds, _ = nds_crd.shape

    tris_conn_subdiv = np.zeros((16*n_tetras, 3), dtype=int)

    for i, tetra10n_conn in enumerate(tetras10n_conn):
        j = 16*i
        n0, n1, n2, n3, n4, n5, n6, n7, n8, n9 = tetra10n_conn

        # triangles connectivity
        tris_conn_subdiv[j] = np.array([n0, n4, n6])
        tris_conn_subdiv[j+1] = np.array([n1, n5, n4])
        tris_conn_subdiv[j+2] = np.array([n4, n5, n6])
        tris_conn_subdiv[j+3] = np.array([n2, n6, n5])

        tris_conn_subdiv[j+4] = np.array([n0, n7, n4])
        tris_conn_subdiv[j+5] = np.array([n4, n7, n8])
        tris_conn_subdiv[j+6] = np.array([n1, n4, n8])
        tris_conn_subdiv[j+7] = np.array([n3, n8, n7])

        tris_conn_subdiv[j+8] = np.array([n0, n6, n7])
        tris_conn_subdiv[j+9] = np.array([n6, n9, n7])
        tris_conn_subdiv[j+10] = np.array([n2, n9, n6])
        tris_conn_subdiv[j+11] = np.array([n3, n7, n9])

        tris_conn_subdiv[j+12] = np.array([n1, n5, n8])
        tris_conn_subdiv[j+13] = np.array([n5, n9, n8])
        tris_conn_subdiv[j+14] = np.array([n2, n9, n5])
        tris_conn_subdiv[j+15] = np.array([n3, n8, n9])

    return tris_conn_subdiv


def tris6n_to_4tris(tris_conn):
    """Get triangles connectivity.

    Six-node triangle is subdivided into four triangles
    """
    n_tris, _ = tris_conn.shape
    # n_nds, _ = nds_crd.shape

    tris_conn_subdiv = np.zeros((4*n_tris, 3), dtype=int)

    for i, tri_conn in enumerate(tris_conn):
        j = 4*i
        n0, n1, n2, n3, n4, n5 = tri_conn

        # triangles connectivity
        tris_conn_subdiv[j  ] = np.array([n0, n3, n5])
        tris_conn_subdiv[j+1] = np.array([n3, n1, n4])
        tris_conn_subdiv[j+2] = np.array([n3, n4, n5])
        tris_conn_subdiv[j+3] = np.array([n5, n4, n2])

    return tris_conn_subdiv



import numpy as np

def split(a, value, source=None):
    # like str.split, but for arrays
    if source is None:
        source = a
    if np.isnan(value):
        idx = np.where(~np.isnan(source))[0]
    else:
        idx = np.where(source != value)[0]

    return np.split(a[idx],np.where(np.diff(idx)!=1)[0]+1)


def join(sep, items):
    return [i for item in items for i in (item, sep)]


def stack(items, sep):
    return [i for item in items for i in (*item, sep)]


if __name__ == "__main__":
    print(join([1,2], [ [[0,0],[0,0]], [[1,1],[1,1]] ]))

    print(stack([ [[0,0],[0,0]], [[1,1],[1,1]] ],  [5,5]))



# Claudio Perez
# Summer 2024
import yaml
import json
import numpy as np
from pathlib import Path
from urllib.parse import urlparse
from scipy.spatial.transform import Rotation

from veux.config import LineStyle, MeshStyle

class State:
    time: float
#   node_array : np.array
#   cell_array : np.array

class so3:
    @classmethod
    def exp(cls, vect):
        return Rotation.from_rotvec(vect).as_matrix()


def read_state(res_file, model=None, only=None, time=None, scale=None, transform=None, recover=None, **opts):

    if hasattr(res_file, "read"):
        res = yaml.load(res_file, Loader=yaml.Loader)

    elif isinstance(res_file, (str,Path)):
        res_path = urlparse(res_file)
        if "json" in res_path[2]:
            with open(res_path[2], "r") as f:
                res = json.loads(f.read())
        else:
            with open(res_path[2], "r") as f:
                res = yaml.load(f, Loader=yaml.Loader)

        if res_path[4]: # query parameters passed
            res = res[int(res_path[4].split("=")[-1])]
    else:
        res = res_file


    #
    # Create the object
    #
#   show_objects = opts.get("show_objects", set())
    recover_rotations = recover # "iter" if "iter" in show_objects else \
                                # "incr" if "incr" in show_objects else None


    if isinstance(res, np.ndarray) or callable(res):
        return BasicState(res, model, transform=transform, scale=scale, time=time)

    # FEDEAS
    elif "IterationHistory" in res or "ConvergedHistory" in res:

        history = StateSeries(res, model,
                    transform =transform,
                    recover_rotations=recover_rotations
                  )

        if recover_rotations is not None:
            history = GroupSeriesSE3(history, model, recover_rotations=recover_rotations, transform=transform)

        if time is not None:
            return history[time]
        else:
            return history


    # Dict of state dicts
    elif isinstance(next(iter(res.values())), dict):
        return {
            k: BasicState(v, model, transform=transform, scale=scale)
                for k,v in res.items()
        }

    # Dict from node tag to nodal values
    else:
        return BasicState(res, model, transform=transform, scale=scale)



class BasicState(State):
    # spatial distribution of a solution
    def __init__(self, data, model, scale=1.0, fn=None, transform=None, time=None):
        """
        data: callable|dict|array
        """
        self.model  = model
        self.time   = time
        self.position   = slice(0, 3)
        self.line_style = LineStyle()
        self.mesh_style = MeshStyle()

        # Handle data
        if isinstance(data, np.ndarray):
            self.ndf = data.shape[1] if len(data.shape) > 1 else 1
            data = {tag: data[i]  for i, tag in enumerate(model.iter_node_tags())}
        else:
            self.ndf = model.ndf

        if self.ndf == 2 or (self.model.ndm == 3 == self.ndf):
            self.rotation = None
        elif self.model.ndm == 2:
            self.rotation = 2
        elif self.model.ndm == 3:
            self.rotation = slice(3, None)

        if callable(data): # OpenSees
            # fn is something like nodeDisp or nodeEigenvector
            data = {tag: data(tag) for tag in model.iter_node_tags()}

        # Handle remaining arguments; by now data should be a dict
        if scale is not None:
            data = {k: scale*np.array(val) for k, val in data.items()}

        if transform is not None:
            data = {k: transform@val for k, val in data.items()}

        self._data  : dict  = data
        self._scale : float = scale

    def __repr__(self):
        return f"<BasicState {self._data}>"


    def node_array(self, tag=None, dof=None):
        # ordered by dof
        if tag is None:
            if dof is None:
                return np.array([self._data[tag] for tag in self.model.iter_node_tags()])
            else:
                return np.array([self._data[tag][dof] for tag in self.model.iter_node_tags()])

        elif dof is None:
            return self._data[tag]

        else:
            return self._data[tag][dof]


    def cell_array(self, tag,
                    dof: int  =None,
                    loc: float=None,
                    node=None,
                    incr=None):

        # incr: init,conv,iter

        return np.array([
                u for n in self.model.cell_nodes(tag)
                    for u in self.node_array(n, dof)
        ])
        return np.array([
                [u for u in self.node_array(n, dof)]
                for n in self.model.cell_nodes(tag)
        ])


class Series:
    @property
    def times(self):
        return np.array(list(self._hist.keys()))

    def node_array(self, tag, dof):
        return np.array([self[time].node_array(tag, dof) for time in self.times])

    def __getitem__(self, time):
        if time < 0:
            return list(self._hist.values())[time]["converged"]
        else:
            return self._hist[time]["converged"]


class GroupSeriesSE3(Series):
    def __init__(self, series: "StateSeries", model, transform=None, **kwds):
        self.model  = model
        if transform is None:
            transform = np.eye(6)

        rotation = GroupSeriesSO3(series, model, transform=transform[3:,3:], **kwds)
        position = series

        self._hist = {
            time: {
                "converged": GroupStateSE3((position[time], rotation[time]), model, time=time)
            } for time in series.times
        }


class GroupStateSE3(State):
    def __init__(self, data, model, time=None):
        self.model  = model
        self._data  = data
        self.time   = time
        self.position = 10
        self.rotation = 20
        self.line_style = LineStyle()
        self.mesh_style = MeshStyle()

        if isinstance(data, tuple):
            self._position = data[0]
            self._rotation = data[1]


    def cell_array(self, tag, dof=None):
        # Note: cant return numpy array, because
        # rotation is different from position
        return [
                self.node_array(node, dof)
                for node in self.model.cell_nodes(tag)
        ]

    def node_array(self, tag, dof):
        if dof == self.position:
            return  self._position.node_array(tag, slice(0,3))
        elif dof == self.rotation:
            return  self._rotation.node_array(tag)
        else:
            return (self._position.node_array(tag),
                    self._rotation.node_array(tag))


class GroupStateSO3(State):
    def __init__(self, data, model, time=None, transform=None):
        self.model  = model
        self.time   = time

        self._data  = data

        if transform is None:
            transform = np.eye(3)
        self._R0 = Rotation.from_matrix(transform)

    def cell_array(self, tag=None):
        if tag is None:
            return np.array([self.cell_array(tag) for tag in self.model.iter_node_tags()])

        Rref = self.model.frame_orientation(tag).T

        return np.array([
            self.node_array(n)@Rref for n in self.model.cell_nodes(tag)
        ])

    def node_array(self, tag=None):
        if tag is None:
            return np.array([self.node_array(tag) for tag in self.model.iter_node_tags()])

        return self._data.get(tag, self._R0).as_matrix()



class StateSeries(Series): # temporal distribution of states
    def __init__(self, soln: "FedeasPost", model, transform=None, scale=None,
                       recover_rotations=None):

        self.model  = model

        self._elements  = model["assembly"]
        self._locations = None
        self._rotations = None
        self._have_iter = "IterationHistory" in soln
        self._recover_rotations = recover_rotations

        # Create history

        self._hist = {}
        time = None
        last = None
        history = soln.get("IterationHistory", soln["ConvergedHistory"])
        i = 0
        for s in history:
            if time != s["Time"]:
                # Store last converged state
                if time is not None:
                    self._hist[time].update({
                        "converged": BasicState(last["U"],  model, transform=transform, scale=scale, time=time),
                        "increment": BasicState(last["DU"], model, transform=transform, scale=scale, time=time),
                    })

                # move time forward
                time = s["Time"]

                # Initialize iteration at new time
                i    = 0
                self._hist[time] = {
                    "iterations": {}
                }
                if "DDU" in s:
                    self._hist[time]["iterations"][i] = BasicState(s["DDU"], model, transform=transform, scale=scale, time=time)

            else:
                i += 1
                self._hist[time]["iterations"][i] = \
                    BasicState(s["DDU"], model, transform=transform, scale=scale, time=time)

            last = s

        time = s["Time"]
        self._hist[time].update({
            "converged": BasicState(last["U"],  model, transform=transform, scale=scale, time=time),
            "increment": BasicState(last["DU"], model, transform=transform, scale=scale, time=time),
        })


    def values(self, incr=None):
        for time in self.times:

            if incr is None:
                yield self[time]

            elif incr == "iter":
                for state in self._hist[time]["iterations"].values():
                    yield state

            elif incr=="conv":
                yield self._hist[time]["increment"]


class GroupSeriesSO3(Series):
    def __init__(self, series: "StateSeries", model, recover_rotations="init", transform=None):
        self.model  = model
        self._series = series
        #
        # Reconstruct group
        #

        if transform is None:
            transform = np.eye(3)

        R0   = Rotation.from_matrix(transform)
        last = {n: R0 for n in self.model.iter_node_tags()}
        hist = self._hist = {}
        time = None
        for incr in series.values(incr=recover_rotations):
            if time != incr.time:
                if time is not None:
                    # Save the converged state at last time
                    hist[time] = {
                        "converged": GroupStateSO3(last, model, time=time)
                    }

                time = incr.time

            last = {
                tag: Rotation.from_rotvec(incr.node_array(tag)[3:])*last[tag]
                for tag in self.model.iter_node_tags()
            }

        # Recover the final converged state
        time = incr.time
        last = {
            tag: Rotation.from_rotvec(incr.node_array(tag)[3:])*last[tag]
            for tag in self.model.iter_node_tags()
        }
        hist[time] = {
            "converged": GroupStateSO3(last, model, time=time)
        }


    def __getitem__(self, time)->GroupStateSO3:
        return self._hist[time]["converged"]


